@@ -21,29 +21,20 @@
 #define CONFIG_H
 
 #include <netdb.h>
+#include <net/if.h>
 #include "types.h"
 #include "auth.h"	/* for the username and password sizes */
-
-#define PORTAL_GROUP_TAG_UNKNOWN -1
-
-/* default iSCSI port number */
-#define ISCSI_DEFAULT_PORT 3260
-
-#define CHAP_AUTHENTICATION     1
+#include "list.h"
 
 /* ISIDs now have a typed naming authority in them.  We use an OUI */
 #define DRIVER_ISID_0  0x00
 #define DRIVER_ISID_1  0x02
 #define DRIVER_ISID_2  0x3D
 
-/* default window size */
-#define TCP_WINDOW_SIZE (256 * 1024)
-
 /* number of possible connections per session */
 #define ISCSI_CONN_MAX		1
-
-/* database version control */
-#define IDBM_VERSION		0x05
+/* max len of interface */
+#define ISCSI_MAX_IFACE_LEN	65
 
 /* the following structures store the options set in the config file.
  * a structure is defined for each logically-related group of options.
@@ -141,18 +132,13 @@ struct iscsi_session_operational_config {
 #define CONFIG_DIGEST_PREFER_OFF 3
 
 struct iscsi_sendtargets_config {
-	char address[NI_MAXHOST];
-	int port;
-	int continuous;
-	int send_async_text;
 	int reopen_max;
 	struct iscsi_auth_config auth;
 	struct iscsi_connection_timeout_config conn_timeo;
+	struct iscsi_conn_operational_config iscsi;
 };
 
 struct iscsi_slp_config {
-	char address[NI_MAXHOST];	/* for unicast */
-	int port;		/* for unicast */
 	char *scopes;
 	char *interfaces;	/* for multicast, list of interfaces names,
 				 * "all", or "none" */
@@ -168,8 +154,10 @@ typedef enum iscsi_startup {
 
 typedef enum discovery_type {
 	DISCOVERY_TYPE_SENDTARGETS,
+	DISCOVERY_TYPE_OFFLOAD_SENDTARGETS,
 	DISCOVERY_TYPE_SLP,
 	DISCOVERY_TYPE_ISNS,
+	DISCOVERY_TYPE_STATIC,
 } discovery_type_e;
 
 typedef struct conn_rec {
@@ -184,6 +172,9 @@ typedef struct conn_rec {
 typedef struct session_rec {
 	int					initial_cmdsn;
 	int					reopen_max;
+	int					cmds_max;
+	int					queue_depth;
+	int					initial_login_retry_max;
 	struct iscsi_auth_config		auth;
 	struct iscsi_session_timeout_config	timeo;
 	struct iscsi_error_timeout_config	err_timeo;
@@ -192,22 +183,39 @@ typedef struct session_rec {
 
 #define ISCSI_TRANSPORT_NAME_MAXLEN 16
 
-typedef struct node_rec {
-	int			dbversion;
-	char			name[TARGET_NAME_MAXLEN];
+typedef struct iface_rec {
+	struct list_head	list;
+	/* iscsi iface record name */
+	char			name[ISCSI_MAX_IFACE_LEN];
+	/* network layer iface name (eth0) */
+	char			netdev[IFNAMSIZ];
+	char			ipaddress[NI_MAXHOST];
+	/*
+	 * TODO: we may have to make this bigger and interconnect
+	 * specific for infinniband 
+	 */
+	char			hwaddress[ISCSI_MAX_IFACE_LEN];
 	char			transport_name[ISCSI_TRANSPORT_NAME_MAXLEN];
+} iface_rec_t;
+
+typedef struct node_rec {
+	struct list_head	list;
+	char			name[TARGET_NAME_MAXLEN];
 	int			tpgt;
-	int			active_conn;
 	iscsi_startup_e		startup;
 	session_rec_t		session;
 	conn_rec_t		conn[ISCSI_CONN_MAX];
+	iface_rec_t		iface;
+	discovery_type_e	disc_type;
+	char			disc_address[NI_MAXHOST];
+	int			disc_port;
 } node_rec_t;
 
 typedef struct discovery_rec {
-	int					id;
-	int					dbversion;
-	iscsi_startup_e				startup;
-	discovery_type_e			type;
+	iscsi_startup_e		startup;
+	discovery_type_e	type;
+	char			address[NI_MAXHOST];
+	int			port;
 	union {
 		struct iscsi_sendtargets_config	sendtargets;
 		struct iscsi_slp_config		slp;
