@@ -45,6 +45,7 @@
 #include "log.h"
 #include "util.h"
 #include "iscsi_sysfs.h"
+#include "iscsi_settings.h"
 
 static void __session_mainloop(void *data);
 static void __conn_error_handle(iscsi_session_t*, iscsi_conn_t*);
@@ -129,28 +130,12 @@ void recvpool_put(iscsi_conn_t *conn, void *handle)
 	}
 }
 
-static void
-__session_online_devs(iscsi_session_t *session)
+static void session_online_devs(int host_no, int sid)
 {
-	sysfs_for_each_device(session->hostno, session->id,
+	sysfs_for_each_device(host_no, sid,
 			      set_device_online);
 }
 
-static void
-write_mgmt_rsp(queue_task_t *qtask, mgmt_ipc_err_e err)
-{
-	log_debug(4, "%s: rsp to fd %d", __FUNCTION__,
-		  qtask->u.login.mgmt_ipc_fd);
-	if (qtask->u.login.mgmt_ipc_fd == 0)
-		return;
-
-	qtask->u.login.rsp.err = err;
-	write(qtask->u.login.mgmt_ipc_fd, &qtask->u.login.rsp,
-	      sizeof(qtask->u.login.rsp));
-	close(qtask->u.login.mgmt_ipc_fd);
-	free(qtask);
-}
-
 static conn_login_status_e
 __login_response_status(iscsi_conn_t *conn,
 		      enum iscsi_login_status login_status)
@@ -367,24 +352,65 @@ __session_conn_create(iscsi_session_t *session, int cid)
 	/* connection's timeouts */
 	conn->id = cid;
 	conn->logout_timeout = conn_rec->timeo.logout_timeout;
+	if (!conn->logout_timeout) {
+		log_error("Invalid timeo.logout_timeout. Must be greater "
+			  "than zero. Using default %d.\n",
+			  DEF_LOGOUT_TIMEO);
+		conn->logout_timeout = DEF_LOGOUT_TIMEO;
+	}
+
 	conn->login_timeout = conn_rec->timeo.login_timeout;
+	if (!conn->login_timeout) {
+		log_error("Invalid timeo.login_timeout. Must be greater "
+			  "than zero. Using default %d.\n",
+			  DEF_LOGIN_TIMEO);
+		conn->login_timeout = DEF_LOGIN_TIMEO;
+	}
+
+	/* noop-out setting */
+	conn->noop_out_interval = conn_rec->timeo.noop_out_interval;
+	conn->noop_out_timeout = conn_rec->timeo.noop_out_timeout;
+	if (conn->noop_out_interval && !conn->noop_out_timeout) {
+		log_error("Invalid timeo.noop_out_timeout. Must be greater "
+			  "than zero. Using default %d.\n",
+			  DEF_NOOP_OUT_TIMEO);
+		conn->noop_out_timeout = DEF_NOOP_OUT_TIMEO;
+	}
+
+	if (conn->noop_out_timeout && !conn->noop_out_interval) {
+		log_error("Invalid timeo.noop_out_interval. Must be greater "
+			  "than zero. Using default %d.\n",
+			  DEF_NOOP_OUT_INTERVAL);
+		conn->noop_out_timeout = DEF_NOOP_OUT_INTERVAL;
+	}
+
+	/*
+	 * currently not used (leftover from linux-iscsi which we
+	 * may do one day)
+	 */
 	conn->auth_timeout = conn_rec->timeo.auth_timeout;
 	conn->active_timeout = conn_rec->timeo.active_timeout;
 	conn->idle_timeout = conn_rec->timeo.idle_timeout;
 	conn->ping_timeout = conn_rec->timeo.ping_timeout;
 
-	/* noop-out setting */
-	conn->noop_out_interval = conn_rec->timeo.noop_out_interval;
-	conn->noop_out_timeout = conn_rec->timeo.noop_out_timeout;
-
 	/* operational parameters */
 	conn->max_recv_dlength =
 			__padding(conn_rec->iscsi.MaxRecvDataSegmentLength);
+	if (conn->max_recv_dlength < ISCSI_MIN_MAX_RECV_SEG_LEN ||
+	    conn->max_recv_dlength > ISCSI_MAX_MAX_RECV_SEG_LEN) {
+		log_error("Invalid iscsi.MaxRecvDataSegmentLength. Must be "
+			 "within %u and %u. Setting to %u\n",
+			  ISCSI_MIN_MAX_RECV_SEG_LEN,
+			  ISCSI_MAX_MAX_RECV_SEG_LEN,
+			  DEF_INI_MAX_RECV_SEG_LEN);
+		conn->max_recv_dlength = DEF_INI_MAX_RECV_SEG_LEN;
+	}
+
 	/*
 	 * iSCSI default, unless declared otherwise by the
 	 * target during login
 	 */
-	conn->max_xmit_dlength = DEFAULT_MAX_RECV_DATA_SEGMENT_LENGTH;
+	conn->max_xmit_dlength = ISCSI_DEF_MAX_RECV_SEG_LEN;
 	conn->hdrdgst_en = conn_rec->iscsi.HeaderDigest;
 	conn->datadgst_en = conn_rec->iscsi.DataDigest;
 
@@ -444,24 +470,33 @@ __send_pdu_timer_remove(struct iscsi_conn *conn)
 
 
 static void
-session_conn_destroy(iscsi_session_t *session, int cid)
+session_stop_conn_timers(iscsi_session_t *session, int cid)
 {
 	iscsi_conn_t *conn = &session->conn[cid];
 
 	__send_pdu_timer_remove(conn);
 	actor_delete(&conn->connect_timer);
-	__recvpool_free(conn);
+}
+
+static void
+session_release(iscsi_session_t *session)
+{
+	log_debug(2, "Releasing session %p", session);
+
+	if (session->target_alias)
+		free(session->target_alias);
+	__recvpool_free(&session->conn[0]);
+	actor_delete(&session->mainloop);
+	queue_destroy(session->queue);
+	free(session);
 }
 
 static void
 session_put(iscsi_session_t *session)
 {
 	session->refcount--;
-	if (session->refcount == 0) {
-		actor_delete(&session->mainloop);
-		queue_destroy(session->queue);
-		free(session);
-	}
+	if (session->refcount == 0)
+		session_release(session);
 }
 
 static void
@@ -471,7 +506,7 @@ session_get(iscsi_session_t *session)
 }
 
 static iscsi_session_t*
-__session_create(node_rec_t *rec, iscsi_provider_t *provider)
+__session_create(node_rec_t *rec, struct iscsi_transport *t)
 {
 	iscsi_session_t *session;
 
@@ -480,12 +515,14 @@ __session_create(node_rec_t *rec, iscsi_provider_t *provider)
 		log_debug(1, "can not allocate memory for session");
 		return NULL;
 	}
+	log_debug(2, "Allocted session %p", session);
 
+	INIT_LIST_HEAD(&session->list);
 	session_get(session);
 	/* opened at daemon load time (iscsid.c) */
 	session->ctrl_fd = control_fd;
-	session->transport_handle = provider->handle;
-	session->provider = provider;
+	session->t = t;
+	session->reopen_qtask.mgmt_ipc_fd = -1;
 
 	/* save node record. we might need it for redirection */
 	memcpy(&session->nrec, rec, sizeof(node_rec_t));
@@ -506,7 +543,27 @@ __session_create(node_rec_t *rec, iscsi_provider_t *provider)
 	session->initial_r2t_en = rec->session.iscsi.InitialR2T;
 	session->imm_data_en = rec->session.iscsi.ImmediateData;
 	session->first_burst = __padding(rec->session.iscsi.FirstBurstLength);
+	if (session->first_burst < ISCSI_MIN_FIRST_BURST_LEN ||
+	    session->first_burst > ISCSI_MAX_FIRST_BURST_LEN) {
+		log_error("Invalid iscsi.FirstBurstLength of %u. Must be "
+			 "within %u and %u. Setting to %u\n",
+			  session->first_burst,
+			  ISCSI_MIN_FIRST_BURST_LEN,
+			  ISCSI_MAX_FIRST_BURST_LEN,
+			  DEF_INI_FIRST_BURST_LEN);
+		session->first_burst = DEF_INI_FIRST_BURST_LEN;
+	}
+
 	session->max_burst = __padding(rec->session.iscsi.MaxBurstLength);
+	if (session->max_burst < ISCSI_MIN_MAX_BURST_LEN ||
+	    session->max_burst > ISCSI_MAX_MAX_BURST_LEN) {
+		log_error("Invalid iscsi.MaxBurstLength of %u. Must be "
+			  "within %u and %u. Setting to %u\n",
+			   session->max_burst, ISCSI_MIN_MAX_BURST_LEN,
+			   ISCSI_MAX_MAX_BURST_LEN, DEF_INI_MAX_BURST_LEN);
+		session->max_burst = DEF_INI_MAX_BURST_LEN;
+	}
+
 	session->def_time2wait = rec->session.iscsi.DefaultTime2Wait;
 	session->def_time2retain = rec->session.iscsi.DefaultTime2Retain;
 	session->erl = rec->session.iscsi.ERL;
@@ -523,7 +580,7 @@ __session_create(node_rec_t *rec, iscsi_provider_t *provider)
 	if (session->replacement_timeout == 0) {
 		log_error("Cannot set replacement_timeout to zero. Setting "
 			  "120 seconds\n");
-		session->replacement_timeout = 120;
+		session->replacement_timeout = DEF_REPLACEMENT_TIMEO;
 	}
 
 	/* OUI and uniqifying number */
@@ -538,28 +595,44 @@ __session_create(node_rec_t *rec, iscsi_provider_t *provider)
 	__setup_authentication(session, &rec->session.auth);
 
 	session->param_mask = 0xFFFFFFFF;
-	if (!(provider->caps & CAP_MULTI_R2T))
+	if (!(t->caps & CAP_MULTI_R2T))
 		session->param_mask &= ~(1 << ISCSI_PARAM_MAX_R2T);
-	if (!(provider->caps & CAP_HDRDGST))
+	if (!(t->caps & CAP_HDRDGST))
 		session->param_mask &= ~(1 << ISCSI_PARAM_HDRDGST_EN);
-	if (!(provider->caps & CAP_DATADGST))
+	if (!(t->caps & CAP_DATADGST))
 		session->param_mask &= ~(1 << ISCSI_PARAM_DATADGST_EN);
-	if (!(provider->caps & CAP_MARKERS)) {
+	if (!(t->caps & CAP_MARKERS)) {
 		session->param_mask &= ~(1 << ISCSI_PARAM_IFMARKER_EN);
 		session->param_mask &= ~(1 << ISCSI_PARAM_OFMARKER_EN);
 	}
 
-	insque(&session->item, &provider->sessions);
-
+	list_add_tail(&session->list, &t->sessions);
 	return session;
 }
 
+static void iscsi_queue_flush(queue_t *queue)
+{
+	unsigned char item_buf[sizeof(queue_item_t) + EVENT_PAYLOAD_MAX];
+	queue_item_t *item = (queue_item_t *)(void *)item_buf;
+
+	/* flush queue by consuming all enqueued items */
+	while (queue_consume(queue, EVENT_PAYLOAD_MAX,
+			     item) != QUEUE_IS_EMPTY) {
+		uintptr_t recv_handle = *(uintptr_t *)queue_item_data(item);
+
+		log_debug(7, "item %p(%d) data size %d flushed", item,
+			  item->event_type, item->data_size);
+		if (item->data_size)
+			recvpool_put(item->context, (void*)recv_handle);
+	}
+}
+
 static void
 __session_destroy(iscsi_session_t *session)
 {
 	log_debug(1, "destroying session\n");
-	remque(&session->item);
-	queue_flush(session->queue);
+	list_del(&session->list);
+	iscsi_queue_flush(session->queue);
 	session_put(session);
 }
 
@@ -580,10 +653,9 @@ session_conn_cleanup(queue_task_t *qtask, mgmt_ipc_err_e err)
 	iscsi_conn_t *conn = qtask->conn;
 	iscsi_session_t *session = conn->session;
 
-	write_mgmt_rsp(qtask, err);
-	session_conn_destroy(session, conn->id);
-	if (conn->id == 0)
-		__session_destroy(session);
+	mgmt_ipc_write_rsp(qtask, err);
+	session_stop_conn_timers(session, conn->id);
+	__session_destroy(session);
 }
 
 static mgmt_ipc_err_e
@@ -594,16 +666,16 @@ __session_conn_shutdown(iscsi_conn_t *conn, queue_task_t *qtask,
 
 	__conn_noop_out_delete(conn);
 	actor_delete(&conn->connect_timer);
-	queue_flush(session->queue);
+	iscsi_queue_flush(session->queue);
 
-	if (ipc->destroy_conn(session->transport_handle, session->id,
+	if (ipc->destroy_conn(session->t->handle, session->id,
 		conn->id)) {
 		log_error("can not safely destroy connection %d", conn->id);
 		return MGMT_IPC_ERR_INTERNAL;
 	}
-	conn->session->provider->utransport->ep_disconnect(conn);
+	conn->session->t->template->ep_disconnect(conn);
 
-	if (ipc->destroy_session(session->transport_handle, session->id)) {
+	if (ipc->destroy_session(session->t->handle, session->id)) {
 		log_error("can not safely destroy session %d", session->id);
 		return MGMT_IPC_ERR_INTERNAL;
 	}
@@ -618,7 +690,7 @@ session_conn_shutdown(iscsi_conn_t *conn, queue_task_t *qtask,
 {
 	iscsi_session_t *session = conn->session;
 
-	if (ipc->stop_conn(session->transport_handle, session->id,
+	if (ipc->stop_conn(session->t->handle, session->id,
 			   conn->id, STOP_CONN_TERM)) {
 		log_error("can't stop connection %d:%d (%d)",
 			  session->id, conn->id, errno);
@@ -668,7 +740,9 @@ __conn_noop_out_timeout(void *data)
 	iscsi_conn_t *conn = (iscsi_conn_t*)data;
 	iscsi_session_t *session = conn->session;
 
-	log_debug(3, "noop out rsp timeout, closing conn...\n");
+	log_warning("Nop-out timedout after %d seconds on connection %d:%d "
+		    "state (%d). Dropping session.", conn->noop_out_timeout,
+		    session->id, conn->id, conn->state);
 	/* XXX: error handle */
 	__conn_error_handle(session, conn);
 }
@@ -698,7 +772,7 @@ __connect_timedout(void *data)
 
 	if (conn->state == STATE_XPT_WAIT) {
 		/* flush any polls or other events queued */
-		queue_flush(session->queue);
+		iscsi_queue_flush(session->queue);
 		log_debug(3, "__connect_timedout queue EV_CONN_TIMER\n");
 		queue_produce(session->queue, EV_CONN_TIMER, qtask, 0, NULL);
 		actor_schedule(&session->mainloop);
@@ -730,7 +804,7 @@ reset_iscsi_params(iscsi_conn_t *conn)
 	 * iSCSI default, unless declared otherwise by the
 	 * target during login
 	 */
-	conn->max_xmit_dlength = DEFAULT_MAX_RECV_DATA_SEGMENT_LENGTH;
+	conn->max_xmit_dlength = ISCSI_DEF_MAX_RECV_SEG_LEN;
 	conn->hdrdgst_en = conn_rec->iscsi.HeaderDigest;
 	conn->datadgst_en = conn_rec->iscsi.DataDigest;
 
@@ -757,7 +831,7 @@ __session_conn_reopen(iscsi_conn_t *conn, queue_task_t *qtask, int do_stop)
 	qtask->conn = conn;
 
 	/* flush stale polls or errors queued */
-	queue_flush(session->queue);
+	iscsi_queue_flush(session->queue);
 	actor_delete(&conn->connect_timer);
 	__conn_noop_out_delete(conn);
 
@@ -766,7 +840,7 @@ __session_conn_reopen(iscsi_conn_t *conn, queue_task_t *qtask, int do_stop)
 
 	if (do_stop) {
 		/* state: STATE_CLEANUP_WAIT */
-		if (ipc->stop_conn(session->transport_handle, session->id,
+		if (ipc->stop_conn(session->t->handle, session->id,
 				   conn->id, do_stop)) {
 			log_error("can't stop connection %d:%d (%d)",
 				  session->id, conn->id, errno);
@@ -775,12 +849,12 @@ __session_conn_reopen(iscsi_conn_t *conn, queue_task_t *qtask, int do_stop)
 		log_debug(3, "connection %d:%d is stopped for recovery",
 			  session->id, conn->id);
 	}
-	conn->session->provider->utransport->ep_disconnect(conn);
+	conn->session->t->template->ep_disconnect(conn);
 
 	if (session->time2wait)
 		goto queue_reopen;
 
-	rc = conn->session->provider->utransport->ep_connect(conn, 1);
+	rc = conn->session->t->template->ep_connect(conn, 1);
 	if (rc < 0 && errno != EINPROGRESS) {
 		char serv[NI_MAXSERV];
 
@@ -840,7 +914,7 @@ iscsi_login_redirect(iscsi_conn_t *conn)
 
 	log_debug(3, "login redirect ...\n");
 
-	queue_flush(session->queue);
+	iscsi_queue_flush(session->queue);
 
 	if (session->r_stage == R_STAGE_NO_CHANGE)
 		session->r_stage = R_STAGE_SESSION_REDIRECT;
@@ -854,38 +928,81 @@ iscsi_login_redirect(iscsi_conn_t *conn)
 }
 
 static void
-print_param_value(enum iscsi_param param, void *value)
+print_param_value(enum iscsi_param param, void *value, int type)
 {
 	log_debug(3, "set operational parameter %d to:", param);
 
-	switch (param) {
-	case ISCSI_PARAM_TARGET_NAME:
-	case ISCSI_PARAM_PERSISTENT_ADDRESS:
-		log_debug(3, "%s", (char *)value);
-		break;
-	default:
+	if (type == ISCSI_STRING)
+		log_debug(3, "%s", value ? (char *)value : "NULL");
+	else
 		log_debug(3, "%u", *(uint32_t *)value);
-		break;
-	}
 }
 
-static void
-__session_scan_host(iscsi_session_t *session, queue_task_t *qtask)
+void free_initiator(void)
+{
+	struct iscsi_transport *t;
+	iscsi_session_t *session, *tmp;
+
+	list_for_each_entry(t, &transports, list) {
+		list_for_each_entry_safe(session, tmp, &t->sessions, list) {
+			list_del(&session->list);
+			session_release(session);
+		}
+	}
+
+	free_transports();
+}
+
+static void session_scan_host(int hostno, queue_task_t *qtask)
 {
 	pid_t pid;
 
-	pid = scan_host(session);
+	pid = scan_host(hostno, 1);
 	if (pid == 0) {
-		write_mgmt_rsp(qtask, MGMT_IPC_OK);
+		mgmt_ipc_write_rsp(qtask, MGMT_IPC_OK);
 		exit(0);
 	} else if (pid > 0) {
-		close(qtask->u.login.mgmt_ipc_fd);
 		need_reap();
-		free(qtask);
+		if (qtask) {
+			close(qtask->mgmt_ipc_fd);
+			free(qtask);
+		}
 	} else
-		write_mgmt_rsp(qtask, MGMT_IPC_ERR_INTERNAL);
+		mgmt_ipc_write_rsp(qtask, MGMT_IPC_ERR_INTERNAL);
 }
 
+static int __iscsi_host_set_param(struct iscsi_transport *t,
+				  int host_no, int param, char *value,
+				  int type)
+{
+	int rc;
+
+	rc = ipc->set_host_param(t->handle, host_no, param, value, type);
+	/* 2.6.20 and below returns EINVAL */
+	if (rc && rc != -ENOSYS && rc != -EINVAL) {
+		log_error("can't set operational parameter %d for "
+			  "host %d, retcode %d (%d)", param, host_no,
+			  rc, errno);
+		return rc;
+	}
+	return 0;
+}
+
+mgmt_ipc_err_e iscsi_host_set_param(int host_no, int param, char *value)
+{
+	struct iscsi_transport *t;
+
+	t = get_transport_by_hba(host_no);
+	if (!t)
+		return MGMT_IPC_ERR_TRANS_FAILURE;
+	if (__iscsi_host_set_param(t, host_no, param, value, ISCSI_STRING))
+		return MGMT_IPC_ERR;
+        return MGMT_IPC_OK;
+}
+
+#define MAX_SESSION_PARAMS 24
+#define MAX_HOST_PARAMS 3
+
 static void
 setup_full_feature_phase(iscsi_conn_t *conn)
 {
@@ -893,114 +1010,152 @@ setup_full_feature_phase(iscsi_conn_t *conn)
 	iscsi_login_context_t *c = &conn->login_context;
 	int i, rc;
 	uint32_t one = 1, zero = 0;
+	struct hostparam {
+		int param;
+		int type;
+		void *value;
+	} hosttbl[MAX_HOST_PARAMS] = {
+		{
+			.param = ISCSI_HOST_PARAM_INITIATOR_NAME,
+			.value = session->initiator_name,
+			.type = ISCSI_STRING,
+		}, {
+			.param = ISCSI_HOST_PARAM_NETDEV_NAME,
+			.value = session->nrec.iface.netdev,
+			.type = ISCSI_STRING,
+		}, {
+			.param = ISCSI_HOST_PARAM_HWADDRESS,
+			.value = session->nrec.iface.hwaddress,
+			.type = ISCSI_STRING,
+		},
+	};
 	struct connparam {
 		int param;
 		int type;
 		void *value;
-		int conn_only; } conntbl[ISCSI_PARAM_SESS_RECOVERY_TMO + 1] = {
-
+		int conn_only;
+	} conntbl[MAX_SESSION_PARAMS] = {
 		{
-		.param = ISCSI_PARAM_MAX_RECV_DLENGTH,
-		.value = &conn->max_recv_dlength,
-		.type = 0,
-		.conn_only = ISCSI_INT,
+			.param = ISCSI_PARAM_MAX_RECV_DLENGTH,
+			.value = &conn->max_recv_dlength,
+			.type = ISCSI_INT,
+			.conn_only = 0,
 		}, {
-		.param = ISCSI_PARAM_MAX_XMIT_DLENGTH,
-		.value = &conn->max_xmit_dlength,
-		.type = 0,
-		.conn_only = ISCSI_INT,
+			.param = ISCSI_PARAM_MAX_XMIT_DLENGTH,
+			.value = &conn->max_xmit_dlength,
+			.type = ISCSI_INT,
+			.conn_only = 0,
 		}, {
-		.param = ISCSI_PARAM_HDRDGST_EN,
-		.value = &conn->hdrdgst_en,
-		.type = 0,
-		.conn_only = ISCSI_INT,
+			.param = ISCSI_PARAM_HDRDGST_EN,
+			.value = &conn->hdrdgst_en,
+			.type = ISCSI_INT,
+			.conn_only = 0,
 		}, {
-		.param = ISCSI_PARAM_DATADGST_EN,
-		.value = &conn->datadgst_en,
-		.type = ISCSI_INT,
-		.conn_only = 1,
+			.param = ISCSI_PARAM_DATADGST_EN,
+			.value = &conn->datadgst_en,
+			.type = ISCSI_INT,
+			.conn_only = 1,
 		}, {
-		.param = ISCSI_PARAM_INITIAL_R2T_EN,
-		.value = &session->initial_r2t_en,
-		.type = ISCSI_INT,
-		.conn_only = 0,
+			.param = ISCSI_PARAM_INITIAL_R2T_EN,
+			.value = &session->initial_r2t_en,
+			.type = ISCSI_INT,
+			.conn_only = 0,
 		}, {
-		.param = ISCSI_PARAM_MAX_R2T,
-		.value = &one, /* FIXME: session->max_r2t */
-		.type = ISCSI_INT,
-		.conn_only = 0,
+			.param = ISCSI_PARAM_MAX_R2T,
+			.value = &one, /* FIXME: session->max_r2t */
+			.type = ISCSI_INT,
+			.conn_only = 0,
 		}, {
-		.param = ISCSI_PARAM_IMM_DATA_EN,
-		.value = &session->imm_data_en,
-		.type = ISCSI_INT,
-		.conn_only = 0,
+			.param = ISCSI_PARAM_IMM_DATA_EN,
+			.value = &session->imm_data_en,
+			.type = ISCSI_INT,
+			.conn_only = 0,
 		}, {
-		.param = ISCSI_PARAM_FIRST_BURST,
-		.value = &session->first_burst,
-		.type = ISCSI_INT,
-		.conn_only = 0,
+			.param = ISCSI_PARAM_FIRST_BURST,
+			.value = &session->first_burst,
+			.type = ISCSI_INT,
+			.conn_only = 0,
 		}, {
-		.param = ISCSI_PARAM_MAX_BURST,
-		.value = &session->max_burst,
-		.type = ISCSI_INT,
-		.conn_only = 0,
+			.param = ISCSI_PARAM_MAX_BURST,
+			.value = &session->max_burst,
+			.type = ISCSI_INT,
+			.conn_only = 0,
 		}, {
-		.param = ISCSI_PARAM_PDU_INORDER_EN,
-		.value = &session->pdu_inorder_en,
-		.type = ISCSI_INT,
-		.conn_only = 0,
+			.param = ISCSI_PARAM_PDU_INORDER_EN,
+			.value = &session->pdu_inorder_en,
+			.type = ISCSI_INT,
+			.conn_only = 0,
 		}, {
-		.param =ISCSI_PARAM_DATASEQ_INORDER_EN,
-		.value = &session->dataseq_inorder_en,
-		.type = ISCSI_INT,
-		.conn_only = 0,
+			.param =ISCSI_PARAM_DATASEQ_INORDER_EN,
+			.value = &session->dataseq_inorder_en,
+			.type = ISCSI_INT,
+			.conn_only = 0,
 		}, {
-		.param = ISCSI_PARAM_ERL,
-		.value = &zero, /* FIXME: session->erl */
-		.type = ISCSI_INT,
-		.conn_only = 0,
+			.param = ISCSI_PARAM_ERL,
+			.value = &zero, /* FIXME: session->erl */
+			.type = ISCSI_INT,
+			.conn_only = 0,
 		}, {
-		.param = ISCSI_PARAM_IFMARKER_EN,
-		.value = &zero,/* FIXME: session->ifmarker_en */
-		.type = ISCSI_INT,
-		.conn_only = 0,
+			.param = ISCSI_PARAM_IFMARKER_EN,
+			.value = &zero,/* FIXME: session->ifmarker_en */
+			.type = ISCSI_INT,
+			.conn_only = 0,
 		}, {
-		.param = ISCSI_PARAM_OFMARKER_EN,
-		.value = &zero,/* FIXME: session->ofmarker_en */
-		.type = ISCSI_INT,
-		.conn_only = 0,
+			.param = ISCSI_PARAM_OFMARKER_EN,
+			.value = &zero,/* FIXME: session->ofmarker_en */
+			.type = ISCSI_INT,
+			.conn_only = 0,
 		}, {
-		.param = ISCSI_PARAM_EXP_STATSN,
-		.value = &conn->exp_statsn,
-		.type = ISCSI_INT,
-		.conn_only = 1,
+			.param = ISCSI_PARAM_EXP_STATSN,
+			.value = &conn->exp_statsn,
+			.type = ISCSI_INT,
+			.conn_only = 1,
 		}, {
-		.param = ISCSI_PARAM_TARGET_NAME,
-		.conn_only = 0,
-		.type = ISCSI_STRING,
-		.value = session->target_name,
+			.param = ISCSI_PARAM_TARGET_NAME,
+			.conn_only = 0,
+			.type = ISCSI_STRING,
+			.value = session->target_name,
 		}, {
-		.param = ISCSI_PARAM_TPGT,
-		.value = &session->portal_group_tag,
-		.type = ISCSI_INT,
-		.conn_only = 0,
+			.param = ISCSI_PARAM_TPGT,
+			.value = &session->portal_group_tag,
+			.type = ISCSI_INT,
+			.conn_only = 0,
 		}, {
-		.param = ISCSI_PARAM_PERSISTENT_ADDRESS,
-		.value = session->nrec.conn[conn->id].address,
-		.type = ISCSI_STRING,
-		.conn_only = 1,
+			.param = ISCSI_PARAM_PERSISTENT_ADDRESS,
+			.value = session->nrec.conn[conn->id].address,
+			.type = ISCSI_STRING,
+			.conn_only = 1,
 		}, {
-		.param = ISCSI_PARAM_PERSISTENT_PORT,
-		.value = &session->nrec.conn[conn->id].port,
-		.type = ISCSI_INT,
-		.conn_only = 1,
+			.param = ISCSI_PARAM_PERSISTENT_PORT,
+			.value = &session->nrec.conn[conn->id].port,
+			.type = ISCSI_INT,
+			.conn_only = 1,
 		}, {
-		.param = ISCSI_PARAM_SESS_RECOVERY_TMO,
-		.value = &session->replacement_timeout,
-		.type = ISCSI_INT,
-		.conn_only = 0,
-		}
-
+			.param = ISCSI_PARAM_SESS_RECOVERY_TMO,
+			.value = &session->replacement_timeout,
+			.type = ISCSI_INT,
+			.conn_only = 0,
+		}, {
+			.param = ISCSI_PARAM_USERNAME,
+			.value = session->username,
+			.type = ISCSI_STRING,
+			.conn_only = 0,
+		}, {
+			.param = ISCSI_PARAM_USERNAME_IN,
+			.value = session->username_in,
+			.type = ISCSI_STRING,
+			.conn_only = 0,
+		}, {
+			.param = ISCSI_PARAM_PASSWORD,
+			.value = session->password,
+			.type = ISCSI_STRING,
+			.conn_only = 0,
+		}, {
+			.param = ISCSI_PARAM_PASSWORD_IN,
+			.value = session->password_in,
+			.type = ISCSI_STRING,
+			.conn_only = 0,
+		},
 		/*
 		 * FIXME: set these timeouts via set_param() API
 		 *
@@ -1025,13 +1180,13 @@ setup_full_feature_phase(iscsi_conn_t *conn)
 	}
 
 	/* Entered full-feature phase! */
-	for (i = 0; i < ISCSI_PARAM_SESS_RECOVERY_TMO + 1; i++) {
+	for (i = 0; i < MAX_SESSION_PARAMS; i++) {
 		if (conn->id != 0 && !conntbl[i].conn_only)
 			continue;
 		if (!(session->param_mask & (1 << conntbl[i].param)))
 			continue;
 
-		rc = ipc->set_param(session->transport_handle, session->id,
+		rc = ipc->set_param(session->t->handle, session->id,
 				   conn->id, conntbl[i].param, conntbl[i].value,
 				   conntbl[i].type);
 		if (rc && rc != -ENOSYS) {
@@ -1045,10 +1200,24 @@ setup_full_feature_phase(iscsi_conn_t *conn)
 			return;
 		}
 
-		print_param_value(conntbl[i].param, conntbl[i].value);
+		print_param_value(conntbl[i].param, conntbl[i].value,
+				  conntbl[i].type);
 	}
 
-	if (ipc->start_conn(session->transport_handle, session->id, conn->id,
+	for (i = 0; i < MAX_HOST_PARAMS; i++) {
+		if (__iscsi_host_set_param(session->t, session->hostno,
+					   hosttbl[i].param, hosttbl[i].value,
+					   hosttbl[i].type)) {
+			__session_conn_shutdown(conn, c->qtask,
+					       MGMT_IPC_ERR_LOGIN_FAILURE);
+			return;
+		}
+
+		print_param_value(hosttbl[i].param, hosttbl[i].value,
+				  hosttbl[i].type);
+	}
+
+	if (ipc->start_conn(session->t->handle, session->id, conn->id,
 			    &rc) || rc) {
 		__session_conn_shutdown(conn, c->qtask,
 				       MGMT_IPC_ERR_INTERNAL);
@@ -1065,13 +1234,15 @@ setup_full_feature_phase(iscsi_conn_t *conn)
 		 * don't want to re-scan it on recovery.
 		 */
 		if (conn->id == 0)
-			__session_scan_host(session, c->qtask);
+			session_scan_host(session->hostno, c->qtask);
 
 		log_warning("connection%d:%d is operational now",
 			    session->id, conn->id);
 	} else {
-		__session_online_devs(session);
-		write_mgmt_rsp(c->qtask, MGMT_IPC_OK);
+		session->sync_qtask = NULL;
+
+		session_online_devs(session->hostno, session->id);
+		mgmt_ipc_write_rsp(c->qtask, MGMT_IPC_OK);
 		log_warning("connection%d:%d is operational after recovery "
 			    "(%d attempts)", session->id, conn->id,
 			     session->reopen_cnt);
@@ -1240,7 +1411,7 @@ __session_conn_recv_pdu(queue_item_t *item)
 	case STATE_LOGOUT_REQUESTED:
 		/* read incomming PDU */
 		if (!iscsi_io_recv_pdu(conn, &hdr, ISCSI_DIGEST_NONE,
-			    conn->data, DEFAULT_MAX_RECV_DATA_SEGMENT_LENGTH,
+			    conn->data, ISCSI_DEF_MAX_RECV_SEG_LEN,
 			    ISCSI_DIGEST_NONE, 0)) {
 			return;
 		}
@@ -1261,45 +1432,21 @@ __session_conn_recv_pdu(queue_item_t *item)
 		}
 		break;
 	case STATE_XPT_WAIT:
-		recvpool_put(conn, conn->recv_handle);
+		recvpool_put(conn, (void *)conn->recv_handle);
 		log_debug(1, "ignoring incomming PDU in XPT_WAIT. "
 			  "let connection re-establish or fail");
 		break;
 	case STATE_CLEANUP_WAIT:
-		recvpool_put(conn, conn->recv_handle);
+		recvpool_put(conn, (void *)conn->recv_handle);
 		log_debug(1, "ignoring incomming PDU in XPT_WAIT. "
 			  "let connection cleanup");
 		break;
 	default:
-		recvpool_put(conn, conn->recv_handle);
+		recvpool_put(conn, (void *)conn->recv_handle);
 		log_error("Invalid state. Dropping PDU.\n");
 	}
 }
 
-static int
-__session_node_established(char *node_name)
-{
-	iscsi_provider_t *p;
-	struct qelem *sitem, *pitem;
-
-	pitem = providers.q_forw;
-	while (pitem != &providers) {
-		p = (iscsi_provider_t *)pitem;
-
-		sitem = p->sessions.q_forw;
-		while (sitem != &p->sessions) {
-			iscsi_session_t *session = (iscsi_session_t *)sitem;
-			if (session->conn[0].state == STATE_LOGGED_IN &&
-			    !strncmp(session->nrec.name, node_name, TARGET_NAME_MAXLEN))
-				return 1;
-			sitem = sitem->q_forw;
-		}
-
-		pitem = pitem->q_forw;
-	}
-	return 0;
-}
-
 static void
 setup_kernel_io_callouts(iscsi_conn_t *conn)
 {
@@ -1325,7 +1472,7 @@ __session_conn_poll(queue_item_t *item)
 	if (conn->state != STATE_XPT_WAIT)
 		return;
 
-	rc = session->provider->utransport->ep_poll(conn, 1);
+	rc = session->t->template->ep_poll(conn, 1);
 	if (rc == 0) {
 		/* timedout: Poll again. */
 		queue_produce(session->queue, EV_CONN_POLL, qtask, 0, NULL);
@@ -1339,8 +1486,10 @@ __session_conn_poll(queue_item_t *item)
 		/* do not allocate new connection in case of reopen */
 		if (session->r_stage == R_STAGE_NO_CHANGE) {
 			if (conn->id == 0 &&
-			    ipc->create_session(session->transport_handle,
+			    ipc->create_session(session->t->handle,
 					session->nrec.session.initial_cmdsn,
+					session->nrec.session.cmds_max,
+					session->nrec.session.queue_depth,
 					&session->id, &session->hostno)) {
 				log_error("can't create session (%d)", errno);
 				err = MGMT_IPC_ERR_INTERNAL;
@@ -1349,15 +1498,7 @@ __session_conn_poll(queue_item_t *item)
 			log_debug(3, "created new iSCSI session %d",
 				  session->id);
 
-			/* unique identifier for OUI */
-			if (__session_node_established(session->nrec.name)) {
-				log_warning("picking unique OUI for "
-					    "the same target node name %s",
-					    session->nrec.name);
-				session->isid[3] = session->id;
-			}
-
-			if (ipc->create_conn(session->transport_handle,
+			if (ipc->create_conn(session->t->handle,
 					session->id, conn->id, &conn->id)) {
 				log_error("can't create connection (%d)",
 					   errno);
@@ -1368,7 +1509,13 @@ __session_conn_poll(queue_item_t *item)
 				  "%d:%d", session->id, conn->id);
 		}
 
-		if (ipc->bind_conn(session->transport_handle, session->id,
+		/*
+		 * TODO: use the iface number or some other value
+		 * so this will be persistent
+		 */
+		session->isid[3] = session->id;
+
+		if (ipc->bind_conn(session->t->handle, session->id,
 				   conn->id, conn->transport_ep_handle,
 				   (conn->id == 0), &rc) || rc) {
 			log_error("can't bind conn %d:%d to session %d, "
@@ -1414,17 +1561,17 @@ __session_conn_poll(queue_item_t *item)
 	return;
 
 c_cleanup:
-	if (ipc->destroy_conn(session->transport_handle, session->id,
+	if (ipc->destroy_conn(session->t->handle, session->id,
                 conn->id)) {
 		log_error("can not safely destroy connection %d:%d",
 			  session->id, conn->id);
 	}
 s_cleanup:
-	if (ipc->destroy_session(session->transport_handle, session->id)) {
+	if (ipc->destroy_session(session->t->handle, session->id)) {
 		log_error("can not safely destroy session %d", session->id);
 	}
 cleanup:
-	session->provider->utransport->ep_disconnect(conn);
+	session->t->template->ep_disconnect(conn);
 	session_conn_cleanup(qtask, err);
 }
 
@@ -1439,7 +1586,7 @@ __session_conn_timer(queue_item_t *item)
 	case STATE_XPT_WAIT:
 		switch (session->r_stage) {
 		case R_STAGE_NO_CHANGE:
-			session->provider->utransport->ep_disconnect(conn);
+			session->t->template->ep_disconnect(conn);
 			log_debug(6, "conn_timer popped at XPT_WAIT: login");
 			/* timeout during initial connect.
 			 * clean connection. write ipc rsp */
@@ -1544,8 +1691,14 @@ __conn_error_handle(iscsi_session_t *session, iscsi_conn_t *conn)
 		break;
 	case STATE_IN_LOGIN:
 		if (session->r_stage == R_STAGE_SESSION_REOPEN) {
-			session_conn_reopen(conn, &session->reopen_qtask,
-					    STOP_CONN_RECOVER);
+			queue_task_t *qtask;
+
+			if (session->sync_qtask)
+				qtask = session->sync_qtask;
+			else
+				qtask = &session->reopen_qtask;
+
+			session_conn_reopen(conn, qtask, STOP_CONN_RECOVER);
 			return;
 		}
 
@@ -1575,15 +1728,16 @@ __conn_error_handle(iscsi_session_t *session, iscsi_conn_t *conn)
 static void
 __session_conn_error(queue_item_t *item)
 {
-	enum iscsi_err error = *(enum iscsi_err *)queue_item_data(item);
+	uintptr_t recv_handle = *(uintptr_t *)queue_item_data(item);
+	enum iscsi_err error = *(enum iscsi_err *)recv_handle;
 	iscsi_conn_t *conn = item->context;
 	iscsi_session_t *session = conn->session;
 
-	log_warning("detected iSCSI connection %d:%d error (%d) "
+	log_warning("Kernel reported iSCSI connection %d:%d error (%d) "
 		    "state (%d)", session->id, conn->id, error,
 		    conn->state);
 	__conn_error_handle(session, conn);
-
+	recvpool_put(conn, (void*)recv_handle);
 }
 
 static void
@@ -1627,80 +1781,36 @@ __session_mainloop(void *data)
 	session_put(session);
 }
 
-static int match_session(void *data, char *targetname, int tpgt, char *address,
-			 int port, int sid)
-{
-	node_rec_t *rec = data;
-	iscsi_provider_t *p;
-
-	log_debug(6, "looking for session [%d][%s,%s,%d]", sid,
-		  rec->name, rec->conn[0].address, rec->conn[0].port);
-
-	p = get_transport_by_sid(sid);
-	if (!p)
-		return 0;
-
-	if (!strcmp(rec->transport_name, p->name) &&
-	    !strcmp(rec->name, targetname) &&
-	    !strcmp(rec->conn[0].address, address) &&
-	    rec->conn[0].port == port)
-		return 1;
-
-	/* keep on looking */
-	return 0;
-}
-
 iscsi_session_t*
 session_find_by_sid(int sid)
 {
-	iscsi_provider_t *p;
+	struct iscsi_transport *t;
 	iscsi_session_t *session;
-	struct qelem *pitem, *sitem;
-
-	pitem = providers.q_forw;
-	while (pitem != &providers) {
-		p = (iscsi_provider_t *)pitem;
-
-		sitem = p->sessions.q_forw;
-		while (sitem != &p->sessions) {
-			session = (iscsi_session_t *)sitem;
 
+	list_for_each_entry(t, &transports, list) {
+		list_for_each_entry(session, &t->sessions, list) {
 			if (session->id == sid)
 				return session;
-			sitem = sitem->q_forw;
 		}
-		pitem = pitem->q_forw;
 	}
-
 	return NULL;
 }
 
 iscsi_session_t*
 session_find_by_rec(node_rec_t *rec)
 {
-	iscsi_provider_t *p;
+	struct iscsi_transport *t;
 	iscsi_session_t *session;
-	struct qelem *pitem, *sitem;
 
-	pitem = providers.q_forw;
-	while (pitem != &providers) {
-		p = (iscsi_provider_t *)pitem;
-
-		sitem = p->sessions.q_forw;
-		while (sitem != &p->sessions) {
-			session = (iscsi_session_t *)sitem;
-
-			if (match_session(rec, session->nrec.name,
-					  -1, session->nrec.conn[0].address,
-					  session->nrec.conn[0].port,
-					  session->id))
+	list_for_each_entry(t, &transports, list) {
+		list_for_each_entry(session, &t->sessions, list) {
+			if (__iscsi_match_session(rec, session->nrec.name,
+					 session->nrec.conn[0].address,
+					 session->nrec.conn[0].port,
+					 &session->nrec.iface))
 				return session;
-
-			sitem = sitem->q_forw;
 		}
-		pitem = pitem->q_forw;
 	}
-
 	return NULL;
 }
 
@@ -1715,7 +1825,7 @@ int session_is_running(node_rec_t *rec)
 	if (session_find_by_rec(rec))
 		return 1;
 
-	if (sysfs_for_each_session(rec, &nr_found, match_session))
+	if (sysfs_for_each_session(rec, &nr_found, iscsi_match_session))
 		return 1;
 
 	return 0;
@@ -1727,63 +1837,60 @@ session_login_task(node_rec_t *rec, queue_task_t *qtask)
 	int rc;
 	iscsi_session_t *session;
 	iscsi_conn_t *conn;
-	iscsi_provider_t *provider;
+	struct iscsi_transport *t;
 
-	if (!rec->active_conn)
-		return MGMT_IPC_ERR_INVAL;
-
-	provider = get_transport_by_name(rec->transport_name);
-	if (!provider)
+	t = get_transport_by_name(rec->iface.transport_name);
+	if (!t)
 		return MGMT_IPC_ERR_TRANS_NOT_FOUND;
-	if (set_uspace_transport(provider))
+	if (set_transport_template(t))
 		return MGMT_IPC_ERR_TRANS_NOT_FOUND;
 
-	if ((!(provider->caps & CAP_RECOVERY_L0) &&
+	if ((!(t->caps & CAP_RECOVERY_L0) &&
 	     rec->session.iscsi.ERL != 0) ||
-	    (!(provider->caps & CAP_RECOVERY_L1) &&
+	    (!(t->caps & CAP_RECOVERY_L1) &&
 	     rec->session.iscsi.ERL > 1)) {
 		log_error("transport '%s' does not support ERL %d",
-			  provider->name, rec->session.iscsi.ERL);
+			  t->name, rec->session.iscsi.ERL);
 		return MGMT_IPC_ERR_TRANS_CAPS;
 	}
 
-	if (!(provider->caps & CAP_MULTI_R2T) &&
+	if (!(t->caps & CAP_MULTI_R2T) &&
 	    rec->session.iscsi.MaxOutstandingR2T) {
 		log_error("transport '%s' does not support "
-			  "MaxOutstandingR2T %d", provider->name,
+			  "MaxOutstandingR2T %d", t->name,
 			  rec->session.iscsi.MaxOutstandingR2T);
 		return MGMT_IPC_ERR_TRANS_CAPS;
 	}
 
-	if (!(provider->caps & CAP_HDRDGST) &&
+	if (!(t->caps & CAP_HDRDGST) &&
 	    rec->conn[0].iscsi.HeaderDigest) {
 		log_error("transport '%s' does not support "
-			  "HeaderDigest != None", provider->name);
+			  "HeaderDigest != None", t->name);
 		return MGMT_IPC_ERR_TRANS_CAPS;
 	}
 
-	if (!(provider->caps & CAP_DATADGST) &&
+	if (!(t->caps & CAP_DATADGST) &&
 	    rec->conn[0].iscsi.DataDigest) {
 		log_error("transport '%s' does not support "
-			  "DataDigest != None", provider->name);
+			  "DataDigest != None", t->name);
 		return MGMT_IPC_ERR_TRANS_CAPS;
 	}
 
-	if (!(provider->caps & CAP_MARKERS) &&
+	if (!(t->caps & CAP_MARKERS) &&
 	    rec->conn[0].iscsi.IFMarker) {
 		log_error("transport '%s' does not support IFMarker",
-			  provider->name);
+			  t->name);
 		return MGMT_IPC_ERR_TRANS_CAPS;
 	}
 
-	if (!(provider->caps & CAP_MARKERS) &&
+	if (!(t->caps & CAP_MARKERS) &&
 	    rec->conn[0].iscsi.OFMarker) {
 		log_error("transport '%s' does not support OFMarker",
-			  provider->name);
+			  t->name);
 		return MGMT_IPC_ERR_TRANS_CAPS;
 	}
 
-	session = __session_create(rec, provider);
+	session = __session_create(rec, t);
 	if (!session)
 		return MGMT_IPC_ERR_LOGIN_FAILURE;
 
@@ -1797,7 +1904,7 @@ session_login_task(node_rec_t *rec, queue_task_t *qtask)
 	conn = &session->conn[0];
 	qtask->conn = conn;
 
-	rc = session->provider->utransport->ep_connect(conn, 1);
+	rc = session->t->template->ep_connect(conn, 1);
 	if (rc < 0 && errno != EINPROGRESS) {
 		char serv[NI_MAXSERV];
 
@@ -1808,7 +1915,7 @@ session_login_task(node_rec_t *rec, queue_task_t *qtask)
 
 		log_error("cannot make a connection to %s:%s (%d)",
 			 conn->host, serv, errno);
-		session_conn_destroy(session, 0);
+		session_stop_conn_timers(session, 0);
 		__session_destroy(session);
 		return MGMT_IPC_ERR_TRANS_FAILURE;
 	}
@@ -1820,8 +1927,8 @@ session_login_task(node_rec_t *rec, queue_task_t *qtask)
 	actor_timer(&conn->connect_timer, conn->login_timeout*1000,
 		    __connect_timedout, qtask);
 
-	qtask->u.login.rsp.command = MGMT_IPC_SESSION_LOGIN;
-	qtask->u.login.rsp.err = MGMT_IPC_OK;
+	qtask->rsp.command = MGMT_IPC_SESSION_LOGIN;
+	qtask->rsp.err = MGMT_IPC_OK;
 
 	return MGMT_IPC_OK;
 }
@@ -1845,16 +1952,16 @@ mgmt_ipc_err_e
 iscsi_sync_session(node_rec_t *rec, queue_task_t *qtask, uint32_t sid)
 {
 	iscsi_session_t *session;
-	iscsi_provider_t *provider;
+	struct iscsi_transport *t;
 	int err;
 
-	provider = get_transport_by_name(rec->transport_name);
-	if (!provider)
+	t = get_transport_by_name(rec->iface.transport_name);
+	if (!t)
 		return MGMT_IPC_ERR_TRANS_NOT_FOUND;
-	if (set_uspace_transport(provider))
+	if (set_transport_template(t))
 		return MGMT_IPC_ERR_TRANS_NOT_FOUND;
 
-	session = __session_create(rec, provider);
+	session = __session_create(rec, t);
 	if (!session)
 		return MGMT_IPC_ERR_NOMEM;
 
@@ -1879,7 +1986,8 @@ iscsi_sync_session(node_rec_t *rec, queue_task_t *qtask, uint32_t sid)
 		goto destroy_session;
 	}
 
-	qtask->u.login.rsp.command = MGMT_IPC_SESSION_SYNC;
+	session->sync_qtask = qtask;
+	qtask->rsp.command = MGMT_IPC_SESSION_SYNC;
 
 	session_conn_reopen(&session->conn[0], qtask, STOP_CONN_RECOVER);
 	log_debug(3, "Started sync iSCSI session %d", session->id);
@@ -1898,9 +2006,10 @@ session_logout_task(iscsi_session_t *session, queue_task_t *qtask)
 	mgmt_ipc_err_e rc = MGMT_IPC_OK;
 
 	conn = &session->conn[0];
-	if (conn->state == STATE_XPT_WAIT &&
+	if (session->sync_qtask ||
+	    (conn->state == STATE_XPT_WAIT &&
 	    (session->r_stage == R_STAGE_NO_CHANGE ||
-	     session->r_stage == R_STAGE_SESSION_REDIRECT)) {
+	     session->r_stage == R_STAGE_SESSION_REDIRECT))) {
 		log_error("session in invalid state for logout. "
 			   "Try again later\n");
 		return MGMT_IPC_ERR_INTERNAL;
@@ -1911,7 +2020,7 @@ session_logout_task(iscsi_session_t *session, queue_task_t *qtask)
 	/* FIXME: implement Logout Request */
 
 	qtask->conn = conn;
-	qtask->u.login.rsp.command = MGMT_IPC_SESSION_LOGOUT;
+	qtask->rsp.command = MGMT_IPC_SESSION_LOGOUT;
 	conn->logout_qtask = qtask;
 
 	switch (conn->state) {
@@ -1933,3 +2042,42 @@ session_logout_task(iscsi_session_t *session, queue_task_t *qtask)
 
 	return rc;
 }
+
+mgmt_ipc_err_e
+iscsi_host_send_targets(queue_task_t *qtask, int host_no, int do_login,
+			struct sockaddr_storage *ss)
+{
+	struct iscsi_transport *t;
+
+	t = get_transport_by_hba(host_no);
+	if (!t || set_transport_template(t)) {
+		log_error("Invalid host no %d for sendtargets\n", host_no);
+		return MGMT_IPC_ERR_TRANS_FAILURE;
+	}
+	if (!(t->caps & CAP_SENDTARGETS_OFFLOAD))
+		return MGMT_IPC_ERR_TRANS_CAPS;
+
+	if (ipc->sendtargets(t->handle, host_no, (struct sockaddr *)ss))
+		return MGMT_IPC_ERR;
+
+	return MGMT_IPC_OK;
+}
+
+/*
+ * HW drivers like qla4xxx present a interface that hides most of the iscsi
+ * details. Userspace sends down a discovery event then it gets notified
+ * if the sessions that were logged in as a result asynchronously, or
+ * the card will have sessions preset in the FLASH and will log into them
+ * automaotically then send us notification that a session is setup.
+ */
+void iscsi_async_session_creation(uint32_t host_no, uint32_t sid)
+{
+	log_debug(3, "session created sid %u host no %d", sid, host_no);
+	session_online_devs(host_no, sid);
+	session_scan_host(host_no, NULL);
+}
+
+void iscsi_async_session_destruction(uint32_t host_no, uint32_t sid)
+{
+	log_debug(3, "session destroyed sid %u host no %d", sid, host_no);
+}
