@@ -21,6 +21,7 @@
 #define INITIATOR_H
 
 #include <stdint.h>
+#include <net/if.h>
 
 #include "types.h"
 #include "iscsi_proto.h"
@@ -30,18 +31,18 @@
 #include "config.h"
 #include "actor.h"
 #include "queue.h"
+#include "list.h"
+
+#define ISCSI_CONFIG_ROOT	"/etc/iscsi/"
+
+#define CONFIG_FILE		ISCSI_CONFIG_ROOT"iscsid.conf"
+#define INITIATOR_NAME_FILE	ISCSI_CONFIG_ROOT"initiatorname.iscsi"
 
-#define ST_CONFIG_DIR		"/etc/iscsi/send_targets"
-#define NODE_CONFIG_DIR		"/etc/iscsi/nodes"
-#define CONFIG_FILE		"/etc/iscsi/iscsid.conf"
 #define PID_FILE		"/var/run/iscsid.pid"
-#define INITIATOR_NAME_FILE	"/etc/iscsi/initiatorname.iscsi"
 #define LOCK_DIR		"/var/lock/iscsi"
 #define LOCK_FILE		"/var/lock/iscsi/lock"
 #define LOCK_WRITE_FILE		"/var/lock/iscsi/lock.write"
 
-#define DEF_ISCSI_PORT		3260
-
 typedef enum conn_login_status_e {
 	CONN_LOGIN_SUCCESS		= 0,
 	CONN_LOGIN_FAILED		= 1,
@@ -120,7 +121,7 @@ typedef struct iscsi_conn {
 	struct iscsi_session *session;
 	iscsi_login_context_t login_context;
 	struct queue_task *logout_qtask;
-	char data[DEFAULT_MAX_RECV_DATA_SEGMENT_LENGTH];
+	char data[ISCSI_DEF_MAX_RECV_SEG_LEN];
 	char host[NI_MAXHOST];	/* scratch */
 	iscsi_conn_state_e state;
 	actor_t connect_timer;
@@ -186,51 +187,22 @@ typedef struct iscsi_conn {
 
 typedef struct queue_task {
 	iscsi_conn_t *conn;
-	union {
-		/* iSCSI requests originated via IPC */
-		struct ipcreq_login {
-			iscsiadm_req_t req;
-			iscsiadm_rsp_t rsp;
-			int mgmt_ipc_fd;
-		} login;
-		struct ipcreq_logout {
-			iscsiadm_req_t req;
-			iscsiadm_rsp_t rsp;
-			int mgmt_ipc_fd;
-		} logout;
-		/* iSCSI requests originated via CTL */
-		struct ctlreq_recv_pdu {
-		} recv_pdu;
-	} u;
+	iscsiadm_req_t req;
+	iscsiadm_rsp_t rsp;
+	int mgmt_ipc_fd;
 } queue_task_t;
 
-typedef enum iscsi_provider_status_e {
-	PROVIDER_STATUS_UNKNOWN		= 0,
-	PROVIDER_STATUS_OPERATIONAL	= 1,
-	PROVIDER_STATUS_FAILED		= 2,
-} iscsi_provider_status_e;
-
-struct iscsi_uspace_transport;
-
-/* represents data path provider */
-typedef struct iscsi_provider_t {
-	struct qelem list;
-	uint64_t handle;
-	uint32_t caps;
-	iscsi_provider_status_e status;
-	char name[ISCSI_TRANSPORT_NAME_MAXLEN];
-	struct qelem sessions;
-	struct iscsi_uspace_transport *utransport;
-} iscsi_provider_t;
+struct iscsi_transport_template;
+struct iscsi_transport;
 
 /* daemon's session structure */
 typedef struct iscsi_session {
-	struct qelem item; /* must stay at the top */
+	struct list_head list;
 	uint32_t id;
 	uint32_t hostno;
 	int refcount;
-	uint64_t transport_handle;
-	iscsi_provider_t *provider;
+	char netdev[IFNAMSIZ];
+	struct iscsi_transport *t;
 	node_rec_t nrec; /* copy of original Node record in database */
 	unsigned int irrelevant_keys_bitmap;
 	int send_async_text;
@@ -283,6 +255,9 @@ typedef struct iscsi_session {
 	iscsi_session_r_stage_e r_stage;
 	uint32_t replacement_timeout;
 
+	/* sync up fields */
+	queue_task_t *sync_qtask;
+
 	/* session's processing */
 	actor_t mainloop;
 	queue_t *queue;
@@ -290,6 +265,31 @@ typedef struct iscsi_session {
 
 } iscsi_session_t;
 
+/*
+ * TODO: replace session_info and host_info with
+ * node_rec and iface_rec
+ */
+struct session_info {
+	struct list_head list;
+	/* local info */
+	struct iface_rec iface;
+	int sid;
+
+	/* remote info */	
+	char targetname[TARGET_NAME_MAXLEN + 1];
+	int tpgt;
+	char address[NI_MAXHOST + 1];
+	int port;
+	char persistent_address[NI_MAXHOST + 1];
+	int persistent_port;
+};
+
+struct host_info {
+	char iname[TARGET_NAME_MAXLEN + 1];
+	struct iface_rec iface;
+	int host_no;
+};
+
 /* login.c */
 
 #define ISCSI_SESSION_TYPE_NORMAL 0
@@ -367,5 +367,17 @@ extern void* recvpool_get(iscsi_conn_t *conn, int ev_size);
 extern void recvpool_put(iscsi_conn_t *conn, void *handle);
 extern mgmt_ipc_err_e iscsi_sync_session(node_rec_t *rec, queue_task_t
 					 *tsk, uint32_t sid);
+extern mgmt_ipc_err_e iscsi_host_send_targets(queue_task_t *qtask,
+			int host_no, int do_login, struct sockaddr_storage *ss);
+extern mgmt_ipc_err_e iscsi_host_set_param(int host_no, int param, char *value);
+extern void iscsi_async_session_creation(uint32_t host_no, uint32_t sid);
+extern void iscsi_async_session_destruction(uint32_t host_no, uint32_t sid);
+extern void free_initiator(void);
+
+/* isns.c */
+extern int isns_init(void);
+extern void isns_handle(int);
+extern void isns_exit(void);
+extern int isns_dev_attr_query_task(queue_task_t *qtask);
 
 #endif /* INITIATOR_H */
