@@ -28,6 +28,10 @@
 #include <sys/wait.h>
 #include <sys/stat.h>
 #include <sys/un.h>
+#include <sys/types.h>
+#include <sys/socket.h>
+#include <arpa/inet.h>
+#include <ifaddrs.h>
 #include <linux/types.h>
 #include <linux/unistd.h>
 
@@ -35,17 +39,21 @@
 #include "initiator.h"
 #include "transport.h"
 #include "version.h"
+#include "iscsi_sysfs.h"
+#include "list.h"
+#include "iscsi_settings.h"
 
-#define ISCSI_TRANSPORT_DIR "/sys/class/iscsi_transport"
-#define ISCSI_SESSION_DIR "/sys/class/iscsi_session"
-#define ISCSI_CONN_DIR "/sys/class/iscsi_connection"
+#define ISCSI_TRANSPORT_DIR	"/sys/class/iscsi_transport"
+#define ISCSI_SESSION_DIR	"/sys/class/iscsi_session"
+#define ISCSI_CONN_DIR		"/sys/class/iscsi_connection"
+#define ISCSI_HOST_DIR		"/sys/class/iscsi_host"
 
 #define ISCSI_MAX_SYSFS_BUFFER NI_MAXHOST
 
 /* tmp buffer used by sysfs functions */
 static char sysfs_file[PATH_MAX];
-int num_providers = 0;
-struct qelem providers;
+int num_transports = 0;
+LIST_HEAD(transports);
 
 int read_sysfs_file(char *filename, void *value, char *format)
 {
@@ -70,10 +78,14 @@ int read_sysfs_file(char *filename, void *value, char *format)
 	return err;
 }
 
-void init_providers(void)
+void free_transports(void)
 {
-	providers.q_forw = &providers;
-	providers.q_back = &providers;
+	struct iscsi_transport *t, *tmp;
+
+	list_for_each_entry_safe(t, tmp, &transports, list) {
+		list_del(&t->list);
+		free(t);
+	}
 }
 
 static int trans_filter(const struct dirent *dir)
@@ -83,17 +95,13 @@ static int trans_filter(const struct dirent *dir)
 
 static int read_transports(void)
 {
-	struct qelem *item;
 	struct dirent **namelist;
 	char filename[64];
 	int i, n, found, err = 0;
-	iscsi_provider_t *p;
+	struct iscsi_transport *t;
 
 	log_debug(7, "in %s", __FUNCTION__);
 
-	if (num_providers == 0)
-		init_providers();
-
 	n = scandir(ISCSI_TRANSPORT_DIR, &namelist, trans_filter,
 		    versionsort);
 	if (n < 0) {
@@ -104,61 +112,67 @@ static int read_transports(void)
 	for (i = 0; i < n; i++) {
 		found = 0;
 
-		/* copy existing pr vider to new array */
-		item = providers.q_forw;
-		while (item != &providers) {
-			p = (iscsi_provider_t *)item;
-
-			if (!strcmp(p->name, namelist[i]->d_name)) {
+		list_for_each_entry(t, &transports, list) {
+			if (!strcmp(t->name, namelist[i]->d_name)) {
 				found = 1;
 				break;
 			}
-			item = item->q_forw;
 		}
 
 		if (found)
 			continue;
 
-		/* copy new provider */
-		p = malloc(sizeof(iscsi_provider_t));
-		if (!p)
+		/* copy new transport */
+		t = malloc(sizeof(*t));
+		if (!t)
 			continue;
+		log_debug(7, "Adding new transport %s", namelist[i]->d_name);
 
-		p->sessions.q_forw = &p->sessions;
-		p->sessions.q_back = &p->sessions;
-
-		strncpy(p->name, namelist[i]->d_name,
+		INIT_LIST_HEAD(&t->sessions);
+		INIT_LIST_HEAD(&t->list);
+		strncpy(t->name, namelist[i]->d_name,
 			ISCSI_TRANSPORT_NAME_MAXLEN);
 
-		sprintf(filename, ISCSI_TRANSPORT_DIR"/%s/handle", p->name);
-		err = read_sysfs_file(filename, &p->handle, "%llu\n");
+		sprintf(filename, ISCSI_TRANSPORT_DIR"/%s/handle", t->name);
+		err = read_sysfs_file(filename, &t->handle, "%llu\n");
 		if (err)
 			continue;
 
-		sprintf(filename, ISCSI_TRANSPORT_DIR"/%s/caps", p->name);
-		err = read_sysfs_file(filename, &p->caps, "0x%x");
+		sprintf(filename, ISCSI_TRANSPORT_DIR"/%s/caps", t->name);
+		err = read_sysfs_file(filename, &t->caps, "0x%x");
 		if (err)
 			continue;
 
-		insque(&p->list, &providers);
+		/*
+		 * tmp hack for qla4xx compat
+		 */
+		if (!strcmp(t->name, "qla4xxx")) {
+			t->caps |= CAP_DATA_PATH_OFFLOAD;
+			t->caps |= CAP_FW_DB;
+		}
+
+		list_add_tail(&t->list, &transports);
 	}
 
 	for (i = 0; i < n; i++)
 		free(namelist[i]);
 	free(namelist);
-	num_providers = n;
+	num_transports = n;
 
 	return 0;
 }
 
-static void get_negotiated_session_param(int sid, char *param, int *value)
+static void get_session_param(int sid, char *param, void *value, char *format)
 {
 	/* set to invalid */
-	*value = -1;
+	if (!strcmp(format, "%s\n"))
+		((char *)value)[0] = '\0';
+	else
+		*((int *)value) = -1;
 
 	memset(sysfs_file, 0, PATH_MAX);
 	sprintf(sysfs_file, ISCSI_SESSION_DIR"/session%d/%s", sid, param);
-	read_sysfs_file(sysfs_file, value, "%d\n");
+	read_sysfs_file(sysfs_file, value, format);
 }
 
 static void get_negotiated_conn_param(int sid, char *param, int *value)
@@ -171,6 +185,21 @@ static void get_negotiated_conn_param(int sid, char *param, int *value)
 	read_sysfs_file(sysfs_file, value, "%d\n");
 }
 
+/* caller must check lengths */
+void get_auth_conf(int sid, struct iscsi_auth_config *conf)
+{
+	memset(conf, 0, sizeof(*conf));
+
+	get_session_param(sid, "username", conf->username, "%s\n");
+	get_session_param(sid, "username_in", conf->username_in, "%s\n");
+	get_session_param(sid, "password", conf->password, "%s\n");
+	if (strlen((char *)conf->password))
+		conf->password_length = strlen((char *)conf->password);
+	get_session_param(sid, "password_in", conf->password_in, "%s\n");
+	if (strlen((char *)conf->password_in))
+		conf->password_in_length = strlen((char *)conf->password_in);
+}
+
 /* called must check for -1=invalid value */
 void get_negotiated_conn_conf(int sid,
 			      struct iscsi_conn_operational_config *conf)
@@ -193,124 +222,221 @@ void get_negotiated_session_conf(int sid,
 {
 	memset(conf, 0, sizeof(*conf));
 
-	get_negotiated_session_param(sid, "data_pdu_in_order",
-				     &conf->DataPDUInOrder);
-	get_negotiated_session_param(sid, "data_seq_in_order",
-				     &conf->DataSequenceInOrder);
-	get_negotiated_session_param(sid, "erl",
-				     &conf->ERL);
-	get_negotiated_session_param(sid, "first_burst_len",
-				     &conf->FirstBurstLength);
-	get_negotiated_session_param(sid, "max_burst_len",
-				     &conf->MaxBurstLength);
-	get_negotiated_session_param(sid, "immediate_data",
-				     &conf->ImmediateData);
-	get_negotiated_session_param(sid, "initial_r2t",
-				     &conf->InitialR2T);
-	get_negotiated_session_param(sid, "max_outstanding_r2t",
-				     &conf->MaxOutstandingR2T);
+	get_session_param(sid, "data_pdu_in_order",
+			  &conf->DataPDUInOrder, "%d\n");
+	get_session_param(sid, "data_seq_in_order",
+			  &conf->DataSequenceInOrder, "%d\n");
+	get_session_param(sid, "erl",
+			  &conf->ERL, "%d\n");
+	get_session_param(sid, "first_burst_len",
+			  &conf->FirstBurstLength, "%d\n");
+	get_session_param(sid, "max_burst_len",
+			  &conf->MaxBurstLength, "%d\n");
+	get_session_param(sid, "immediate_data",
+			  &conf->ImmediateData, "%d\n");
+	get_session_param(sid, "initial_r2t",
+			  &conf->InitialR2T, "%d\n");
+	get_session_param(sid, "max_outstanding_r2t",
+			  &conf->MaxOutstandingR2T, "%d\n");
 }
 
-int get_sessioninfo_by_sysfs_id(int *sid, char *targetname, char *addr,
-				int *port, int *tpgt, char *session)
+uint32_t get_host_no_from_sid(uint32_t sid, int *err)
 {
-	int ret;
+	char *buf, *path, *tmp;
+	uint32_t host_no;
 
-	if (sscanf(session, "session%d", sid) != 1) {
-		log_error("invalid session '%s'", session);
-		return EINVAL;
+	*err = 0;
+
+	buf = calloc(2, PATH_MAX);
+	if (!buf) {
+		*err = ENOMEM;
+		return 0;
+	}
+	path = buf + PATH_MAX;
+
+	sprintf(path, ISCSI_SESSION_DIR"/session%d/device", sid);
+	if (readlink(path, buf, PATH_MAX) < 0) {
+		log_error("Could not get link for %s.", path);
+		*err = errno;
+		goto free_buf;
 	}
 
-	memset(sysfs_file, 0, PATH_MAX);
-	sprintf(sysfs_file, ISCSI_SESSION_DIR"/%s/targetname", session);
-	ret = read_sysfs_file(sysfs_file, targetname, "%s\n");
-	if (ret) {
-		log_error("could not read session targetname: %d", ret);
-		return ret;
+	/* buf will be .....bus_info/hostX/sessionY */
+
+	/* find hostX */
+	tmp = strrchr(buf, '/');
+	*tmp = '\0';
+
+	/* find bus and increment past it */
+	tmp = strrchr(buf, '/');
+	tmp++;
+
+	if (sscanf(tmp, "host%u", &host_no) != 1) {
+		log_error("Could not get host for sid %u.", sid);
+		*err = ENXIO;
+		goto free_buf;
 	}
 
-	memset(sysfs_file, 0, PATH_MAX);
-	sprintf(sysfs_file, ISCSI_SESSION_DIR"/%s/tpgt", session);
-	ret = read_sysfs_file(sysfs_file, tpgt, "%u\n");
-	if (ret) {
-		log_error("could not read session tpgt: %d", ret);
-		return ret;
-	}
+free_buf:
+	free(buf);
+	return host_no;
+}
 
-	/* some HW drivers do not export addr and port */
-	memset(sysfs_file, 0, PATH_MAX);
-	sprintf(sysfs_file, ISCSI_CONN_DIR"/connection%d:0/"
-		"persistent_address", *sid);
-	memset(addr, 0, NI_MAXHOST);
-	ret = read_sysfs_file(sysfs_file, addr, "%s\n");
-	if (ret) {
-		/* fall back to current address */
-		log_debug(5, "could not read pers conn addr: %d", ret);
-		memset(sysfs_file, 0, PATH_MAX);
-		sprintf(sysfs_file, ISCSI_CONN_DIR"/connection%d:0/address",
-			 *sid);
-		memset(addr, 0, NI_MAXHOST);
-		ret = read_sysfs_file(sysfs_file, addr, "%s\n");
-		if (ret)
-			log_debug(5, "could not read curr addr: %d", ret);
-	}
+static int __get_host_no_from_hwaddress(void *data, struct host_info *info)
+{
+	struct host_info *ret_info = data;
 
-	memset(sysfs_file, 0, PATH_MAX);
-	sprintf(sysfs_file, ISCSI_CONN_DIR"/connection%d:0/"
-		"persistent_port", *sid);
-	*port = -1;
-	ret = read_sysfs_file(sysfs_file, port, "%u\n");
-	if (ret) {
-		/* fall back to current port */
-		log_debug(5, "Could not read pers conn port %d\n", ret);
-		memset(sysfs_file, 0, PATH_MAX);
-		sprintf(sysfs_file, ISCSI_CONN_DIR"/connection%d:0/port",
-			*sid);
-		*port = -1;
-		ret = read_sysfs_file(sysfs_file, port, "%u\n");
-		if (ret)
-			log_debug(5, "Could not read curr conn port %d\n", ret);
+	if (!strcmp(ret_info->iface.hwaddress, info->iface.hwaddress)) {
+		ret_info->host_no = info->host_no;
+		return 1;
 	}
-
-	log_debug(7, "found targetname %s address %s port %d\n",
-		  targetname, addr ? addr : "NA", *port);
 	return 0;
 }
 
-int sysfs_for_each_session(void *data, int *nr_found,
-			   int (* fn)(void *, char *, int, char *, int, int))
+static uint32_t get_host_no_from_hwaddress(char *address, int *rc)
+{
+	uint32_t host_no = -1;
+	struct host_info *info;
+	int nr_found, local_rc;
+
+	*rc = 0;
+
+	info = calloc(1, sizeof(*info));
+	if (!info) {
+		*rc = ENOMEM;
+		return -1;
+	}
+	strcpy(info->iface.hwaddress, address);
+
+	local_rc = sysfs_for_each_host(info, &nr_found,
+					__get_host_no_from_hwaddress);
+	if (local_rc == 1)
+		host_no = info->host_no;
+	else
+		*rc = ENODEV;
+	free(info);
+	return host_no;
+}
+
+static int __get_host_no_from_ipaddress(void *data, struct host_info *info)
+{
+	struct host_info *ret_info = data;
+
+	if (!strcmp(ret_info->iface.ipaddress, info->iface.ipaddress)) {
+		ret_info->host_no = info->host_no;
+		return 1;
+	}
+	return 0;
+}
+
+static uint32_t get_host_no_from_ipaddress(char *address, int *rc)
+{
+	uint32_t host_no = -1;
+	struct host_info *info;
+	int nr_found;
+	int local_rc;
+
+	*rc = 0;
+
+	info = calloc(1, sizeof(*info));
+	if (!info) {
+		*rc = ENOMEM;
+		return -1;
+	}
+	strcpy(info->iface.ipaddress, address);
+
+	local_rc = sysfs_for_each_host(info, &nr_found,
+					__get_host_no_from_ipaddress);
+	if (local_rc == 1)
+		host_no = info->host_no;
+	else
+		*rc = ENODEV;
+	free(info);
+	return host_no;
+}
+
+uint32_t get_host_no_from_iface(struct iface_rec *iface, int *rc)
+{
+	int tmp_rc;
+	uint32_t host_no = -1;
+
+	if (strlen(iface->hwaddress) &&
+	    strcasecmp(iface->hwaddress, DEFAULT_HWADDRESS))
+		host_no = get_host_no_from_hwaddress(iface->hwaddress, &tmp_rc);
+	else if (strlen(iface->ipaddress) &&
+		 strcasecmp(iface->ipaddress, DEFAULT_IPADDRESS))
+		host_no = get_host_no_from_ipaddress(iface->ipaddress, &tmp_rc);
+	else
+		tmp_rc = EINVAL;
+
+	*rc = tmp_rc;
+	return host_no;
+}
+
+int sysfs_for_each_host(void *data, int *nr_found, sysfs_host_op_fn *fn)
 {
 	struct dirent **namelist;
-	int rc = 0, sid, port, tpgt, n, i;
-	char *targetname, *address;
+	int rc = 0, i, n;
+	struct host_info *info;
+	struct iscsi_transport *t;
 
-	targetname = malloc(TARGET_NAME_MAXLEN + 1);
-	if (!targetname)
+	info = calloc(1, sizeof(*info));
+	if (!info)
 		return ENOMEM;
 
-	address = malloc(NI_MAXHOST + 1);
-	if (!address) {
-		rc = ENOMEM;
-		goto free_target;
-	}
-
-	sprintf(sysfs_file, ISCSI_SESSION_DIR);
-	n = scandir(sysfs_file, &namelist, trans_filter,
+	n = scandir(ISCSI_HOST_DIR, &namelist, trans_filter,
 		    versionsort);
 	if (n <= 0)
-		goto free_address;
+		goto free_info;
 
 	for (i = 0; i < n; i++) {
-		rc = get_sessioninfo_by_sysfs_id(&sid, targetname, address,
-						 &port, &tpgt,
-						 namelist[i]->d_name);
-		if (rc) {
-			log_error("could not find session info for %s",
-				   namelist[i]->d_name);
-			continue;
+		if (sscanf(namelist[i]->d_name, "host%u", &info->host_no) !=
+			   1) {
+			log_error("Invalid iscsi host dir: %s",
+				  namelist[i]->d_name);
+			break;
 		}
 
-		rc = fn(data, targetname, tpgt, address, port, sid);
+		memset(sysfs_file, 0, PATH_MAX);
+		sprintf(sysfs_file, ISCSI_HOST_DIR"/%s/initiatorname",
+			namelist[i]->d_name);
+		rc = read_sysfs_file(sysfs_file, info->iname, "%s\n");
+		if (rc)
+			log_debug(4, "Could not read initiatorname for host "
+				  "%u. Error %d\n", info->host_no, rc);
+
+		memset(sysfs_file, 0, PATH_MAX);
+		sprintf(sysfs_file, ISCSI_HOST_DIR"/%s/ipaddress",
+			namelist[i]->d_name);
+		rc = read_sysfs_file(sysfs_file, info->iface.ipaddress, "%s\n");
+		if (rc)
+			log_debug(4, "Could not read ipaddress for host %u. "
+				  "Error %d\n", info->host_no, rc);
+
+		memset(sysfs_file, 0, PATH_MAX);
+		sprintf(sysfs_file, ISCSI_HOST_DIR"/%s/hwaddress",
+			namelist[i]->d_name);
+		rc = read_sysfs_file(sysfs_file, info->iface.hwaddress, "%s\n");
+		if (rc)
+			log_debug(4, "Could not read hwaddress for host %u. "
+				  "Error %d\n", info->host_no, rc);
+
+		memset(sysfs_file, 0, PATH_MAX);
+		sprintf(sysfs_file, ISCSI_HOST_DIR"/%s/netdev",
+			namelist[i]->d_name);
+		rc = read_sysfs_file(sysfs_file, info->iface.netdev, "%s\n");
+		if (rc)
+			log_debug(4, "Could not read netdev for host %u. "
+				  "Error %d\n", info->host_no, rc);
+
+		t = get_transport_by_hba(info->host_no);
+		if (!t)
+			log_debug(4, "could not get transport name for host %d",
+				  info->host_no);
+		else
+			strcpy(info->iface.transport_name, t->name);
+
+		rc = fn(data, info);
 		if (rc != 0)
 			break;
 		(*nr_found)++;
@@ -320,10 +446,190 @@ int sysfs_for_each_session(void *data, int *nr_found,
 		free(namelist[i]);
 	free(namelist);
 
-free_address:
-	free(address);
-free_target:
-	free(targetname);
+free_info:
+	free(info);
+	return rc;
+}
+
+int get_sessioninfo_by_sysfs_id(struct session_info *info, char *session)
+{
+	int ret, pers_failed = 0;
+	uint32_t host_no;
+	struct iscsi_transport *t;
+
+	if (sscanf(session, "session%d", &info->sid) != 1) {
+		log_error("invalid session '%s'", session);
+		return EINVAL;
+	}
+
+	memset(sysfs_file, 0, PATH_MAX);
+	sprintf(sysfs_file, ISCSI_SESSION_DIR"/%s/targetname", session);
+	ret = read_sysfs_file(sysfs_file, info->targetname, "%s\n");
+	if (ret) {
+		log_error("could not read session targetname: %d", ret);
+		return ret;
+	}
+
+	memset(sysfs_file, 0, PATH_MAX);
+	sprintf(sysfs_file, ISCSI_SESSION_DIR"/%s/tpgt", session);
+	ret = read_sysfs_file(sysfs_file, &info->tpgt, "%u\n");
+	if (ret) {
+		log_error("could not read session tpgt: %d", ret);
+		return ret;
+	}
+
+	/* some HW drivers do not export addr and port */
+	memset(sysfs_file, 0, PATH_MAX);
+	sprintf(sysfs_file, ISCSI_CONN_DIR"/connection%d:0/"
+		"persistent_address", info->sid);
+	memset(info->persistent_address, 0, NI_MAXHOST);
+	ret = read_sysfs_file(sysfs_file, info->persistent_address, "%s\n");
+	if (ret) {
+		pers_failed = 1;
+		/* older qlogic does not support this */
+		log_debug(5, "could not read pers conn addr: %d", ret);
+	}
+
+	memset(sysfs_file, 0, PATH_MAX);
+	sprintf(sysfs_file, ISCSI_CONN_DIR"/connection%d:0/address",
+		 info->sid);
+	memset(info->address, 0, NI_MAXHOST);
+	ret = read_sysfs_file(sysfs_file, info->address, "%s\n");
+	if (ret) {
+		log_debug(5, "could not read curr addr: %d", ret);
+		/* iser did not export this */
+		if (!pers_failed)
+			strcpy(info->address, info->persistent_address);
+	} else if (pers_failed)
+		/*
+		 * for qla if we could not get the persistent addr
+		 * we will use the current for both addrs
+		 */
+		strcpy(info->persistent_address, info->address);
+	pers_failed = 0;
+
+	memset(sysfs_file, 0, PATH_MAX);
+	sprintf(sysfs_file, ISCSI_CONN_DIR"/connection%d:0/"
+		"persistent_port", info->sid);
+	info->persistent_port = -1;
+	ret = read_sysfs_file(sysfs_file, &info->persistent_port, "%u\n");
+	if (ret) {
+		pers_failed = 1;
+		log_debug(5, "Could not read pers conn port %d", ret);
+	}
+
+	memset(sysfs_file, 0, PATH_MAX);
+	sprintf(sysfs_file, ISCSI_CONN_DIR"/connection%d:0/port",
+		info->sid);
+	info->port = -1;
+	ret = read_sysfs_file(sysfs_file, &info->port, "%u\n");
+	if (ret) {
+		/* iser did not export this */
+		if (!pers_failed)
+			info->port = info->persistent_port;
+		log_debug(5, "Could not read curr conn port %d", ret);
+	} else if (pers_failed)
+		/*
+		 * for qla if we could not get the persistent addr
+		 * we will use the current for both addrs
+		 */
+		info->persistent_port = info->port;
+
+	ret = 0;
+	host_no = get_host_no_from_sid(info->sid, &ret);
+	if (ret) {
+		log_error("could not get host_no for session %d.", ret);
+		return ret;
+	}
+
+	memset(sysfs_file, 0, PATH_MAX);
+	sprintf(sysfs_file, ISCSI_HOST_DIR"/host%u/hwaddress", host_no);
+	/*
+	 * backward compat
+	 * If we cannot get the address we assume we are doing the old
+	 * style and use default.
+	 */
+	sprintf(info->iface.hwaddress, DEFAULT_HWADDRESS);
+	ret = read_sysfs_file(sysfs_file, info->iface.hwaddress, "%s\n");
+	if (ret)
+		log_debug(7, "could not read hwaddress for %s", sysfs_file);
+
+	t = get_transport_by_sid(info->sid);
+	if (!t)
+		log_debug(7, "could not get transport name for session %d",
+			  info->sid);
+	else
+		strcpy(info->iface.transport_name, t->name);
+
+	memset(sysfs_file, 0, PATH_MAX);
+	sprintf(sysfs_file, ISCSI_HOST_DIR"/host%u/ipaddress", host_no);
+	/* if not found just print out default */
+	sprintf(info->iface.ipaddress, DEFAULT_IPADDRESS);
+	ret = read_sysfs_file(sysfs_file, info->iface.ipaddress, "%s\n");
+	if (ret)
+		log_debug(7, "could not read local address for %s",
+			 sysfs_file);
+
+	memset(sysfs_file, 0, PATH_MAX);
+	sprintf(sysfs_file, ISCSI_HOST_DIR"/host%u/netdev", host_no);
+	/* if not found just print out default */
+	sprintf(info->iface.netdev, DEFAULT_NETDEV);
+	ret = read_sysfs_file(sysfs_file, info->iface.netdev, "%s\n");
+	if (ret)
+		log_debug(7, "could not read netdev for %s",
+			 sysfs_file);
+
+	log_debug(7, "found targetname %s address %s pers address %s port %d "
+		 "pers port %d driver %s iface ipaddress %s "
+		 "netdev %s hwaddress %s",
+		  info->targetname, info->address ? info->address : "NA",
+		  info->persistent_address ? info->persistent_address : "NA",
+		  info->port, info->persistent_port,
+		  info->iface.transport_name, info->iface.ipaddress,
+		  info->iface.netdev, info->iface.hwaddress);
+	return 0;
+}
+ 
+int sysfs_for_each_session(void *data, int *nr_found, sysfs_session_op_fn *fn)
+{
+	struct dirent **namelist;
+	int rc = 0, n, i;
+	struct session_info *info;
+
+	info = calloc(1, sizeof(*info));
+	if (!info)
+		return ENOMEM;
+
+	sprintf(sysfs_file, ISCSI_SESSION_DIR);
+	n = scandir(sysfs_file, &namelist, trans_filter,
+		    versionsort);
+	if (n <= 0)
+		goto free_info;
+
+	for (i = 0; i < n; i++) {
+		rc = get_sessioninfo_by_sysfs_id(info, namelist[i]->d_name);
+		if (rc) {
+			log_error("could not find session info for %s",
+				   namelist[i]->d_name);
+			continue;
+		}
+
+		rc = fn(data, info);
+		if (rc > 0)
+			break;
+		else if (rc == 0)
+			(*nr_found)++;
+		else
+			/* if less than zero it means it was not a match */
+			rc = 0;
+	}
+
+	for (i = 0; i < n; i++)
+		free(namelist[i]);
+	free(namelist);
+
+free_info:
+	free(info);
 	return rc;
 }
 
@@ -409,64 +715,24 @@ static uint32_t get_target_no_from_sid(uint32_t sid, int *err)
 
 }
 
-uint32_t get_host_no_from_sid(uint32_t sid, int *err)
+struct iscsi_transport *get_transport_by_name(char *transport_name)
 {
-	char buf[PATH_MAX], path[PATH_MAX], *tmp;
-	uint32_t host_no;
-
-	*err = 0;
-
-	memset(buf, 0, PATH_MAX);
-	memset(path, 0, PATH_MAX);
-	sprintf(path, ISCSI_SESSION_DIR"/session%d/device", sid);
-	if (readlink(path, buf, PATH_MAX) < 0) {
-		log_error("Could not get link for %s\n", path);
-		*err = errno;
-		return 0;
-	}
-
-	/* buf will be .....bus_info/hostX/sessionY */
-
-	/* find hostX */
-	tmp = strrchr(buf, '/');
-	*tmp = '\0';
-
-	/* find bus and increment past it */
-	tmp = strrchr(buf, '/');
-	tmp++;
-
-	if (sscanf(tmp, "host%u", &host_no) != 1) {
-		log_error("Could not get host for sid %u\n", sid);
-		*err = ENXIO;
-		return 0;
-	}
-
-	return host_no;
-}
-
-iscsi_provider_t *get_transport_by_name(char *transport_name)
-{
-	struct qelem *pitem;
-	iscsi_provider_t *p;
+	struct iscsi_transport *t;
 
 	/* sync up kernel and userspace */
 	read_transports();
 
-	/* check if the transport is loaded */
-	pitem = providers.q_forw;
-	while (pitem != &providers) {
-		p = (iscsi_provider_t *)pitem;
-
-		if (p->handle && !strncmp(p->name, transport_name,
+	/* check if the transport is loaded and matches */
+	list_for_each_entry(t, &transports, list) {
+		if (t->handle && !strncmp(t->name, transport_name,
 					  ISCSI_TRANSPORT_NAME_MAXLEN))
-			return p;
-		pitem = pitem->q_forw;
+			return t;
 	}
 	return NULL;
 }
 
 /* TODO: replace the following functions with some decent sysfs links */
-iscsi_provider_t *get_transport_by_hba(long host_no)
+struct iscsi_transport *get_transport_by_hba(long host_no)
 {
 	char name[ISCSI_TRANSPORT_NAME_MAXLEN];
 	int rc;
@@ -478,7 +744,7 @@ iscsi_provider_t *get_transport_by_hba(long host_no)
 	sprintf(sysfs_file, "/sys/class/scsi_host/host%lu/proc_name", host_no);
 	rc = read_sysfs_file(sysfs_file, name, "%s\n");
 	if (rc) {
-		log_error("Could not read %s rc %d\n", sysfs_file, rc);
+		log_error("Could not read %s rc %d.", sysfs_file, rc);
 		return NULL;
 	}
 
@@ -492,7 +758,7 @@ iscsi_provider_t *get_transport_by_hba(long host_no)
 		return get_transport_by_name(name);
 }
 
-iscsi_provider_t *get_transport_by_sid(uint32_t sid)
+struct iscsi_transport *get_transport_by_sid(uint32_t sid)
 {
 	uint32_t host_no;
 	int err;
@@ -516,7 +782,7 @@ int set_exp_statsn(iscsi_conn_t *conn)
 		ISCSI_CONN_DIR"/connection%d:%d/exp_statsn",
 		conn->session->id, conn->id);
 	if (read_sysfs_file(sysfs_file, &conn->exp_statsn, "%u\n")) {
-		log_error("Could not read %s. Using zero fpr exp_statsn\n",
+		log_error("Could not read %s. Using zero fpr exp_statsn.",
 			  sysfs_file);
 		conn->exp_statsn = 0;
 	}
@@ -566,8 +832,7 @@ void set_device_online(int hostno, int target, int lun)
 	log_debug(4, "online device using %s", sysfs_file);
 	if (write(fd, "running\n", 8) == -1 && errno != EINVAL)
 		/* we should read the state */
-		log_error("Could not online LUN %d err %d\n",
-			  lun, errno);
+		log_error("Could not online LUN %d err %d.", lun, errno);
 	close(fd);
 }
 
@@ -586,7 +851,7 @@ void delete_device(int hostno, int target, int lun)
 	close(fd);
 }
 
-pid_t __scan_host(int hostno, int async)
+pid_t scan_host(int hostno, int async)
 {
 	pid_t pid = 0;
 	int fd;
@@ -595,7 +860,7 @@ pid_t __scan_host(int hostno, int async)
 		hostno);
 	fd = open(sysfs_file, O_WRONLY);
 	if (fd < 0) {
-		log_error("could not scan scsi host%d\n", hostno);
+		log_error("could not scan scsi host%d.", hostno);
 		return -1;
 	}
 
@@ -615,27 +880,19 @@ pid_t __scan_host(int hostno, int async)
 		 * scan by hand
 		  */
 		log_error("Could not start scanning process for host %d "
-			  "err %d. Try scanning through sysfs\n", hostno,
+			  "err %d. Try scanning through sysfs.", hostno,
 			  errno);
 
 	close(fd);
 	return pid;
 }
 
-/*
- * Scan a session from usersapce using sysfs
- */
-pid_t scan_host(iscsi_session_t *session, int async)
-{
-	return __scan_host(session->hostno, 1);
-}
-
-iscsi_provider_t *get_transport_by_session(char *sys_session)
+struct iscsi_transport *get_transport_by_session(char *sys_session)
 {
 	uint32_t sid;
 
         if (sscanf(sys_session, "session%u", &sid) != 1) {
-                log_error("invalid session '%s'", sys_session);
+                log_error("invalid session '%s'.", sys_session);
                 return NULL;
         }
 
@@ -658,7 +915,7 @@ void check_class_version(void)
 	if (get_iscsi_kernel_version(version))
 		goto fail;
 
-	log_warning("transport class version %s. iscsid version %s\n",
+	log_warning("transport class version %s. iscsid version %s",
 		    version, ISCSI_VERSION_STR);
 
 	for (i = 0; i < strlen(version); i++) {
@@ -679,9 +936,9 @@ void check_class_version(void)
 		return;
 
 fail:
-	log_error("Invalid version from %s. Make sure a up to date "
-		  "scsi_transport_iscsi module is loaded and a up to"
-		  "date version of iscsid is running. Exiting...\n",
-		  ISCSI_VERSION_FILE);
+	log_error( "Missing or Invalid version from %s. Make sure a up "
+		"to date scsi_transport_iscsi module is loaded and a up to"
+		"date version of iscsid is running. Exiting...",
+		ISCSI_VERSION_FILE);
 	exit(1);
 }
