@@ -27,13 +27,13 @@ extern char initiator_name[];
 extern char initiator_alias[];
 
 /* discovery.c */
-extern int sendtargets_discovery(struct iscsi_sendtargets_config *config,
-				 struct string_buffer *info);
-extern int slp_discovery(struct iscsi_slp_config *config);
-extern int add_target_record(struct string_buffer *info, char *name, char *end,
-			     int lun_inventory_changed, char *default_address,
-			     char *default_port);
-extern int add_portal(struct string_buffer *info, char *address, char *port,
-		      char *tag);
+struct idbm;
+struct discovery_rec;
+struct list_head;
+
+extern int discovery_sendtargets(struct idbm *db, struct discovery_rec *drec,
+				 struct list_head *ifaces);
+extern int discovery_offload_sendtargets(struct idbm *db, int host_no,
+					 int do_login, discovery_rec_t *drec);
 
 #endif /* ISCSIADM_H */
