@@ -46,6 +46,7 @@
 #include "idbm.h"
 #include "version.h"
 #include "iscsi_sysfs.h"
+#include "iscsi_settings.h"
 
 /* global config info */
 struct iscsi_daemon_config daemon_config;
@@ -53,6 +54,7 @@ struct iscsi_daemon_config *dconfig = &daemon_config;
 
 static char program_name[] = "iscsid";
 int control_fd, mgmt_ipc_fd;
+static pid_t log_pid;
 
 extern char sysfs_file[];
 
@@ -92,42 +94,150 @@ Open-iSCSI initiator daemon.\n\
 	exit(status == 0 ? 0 : -1);
 }
 
-static int sync_session(void *data, char *targetname, int tpgt, char *address,
-			int port, int sid)
+static void
+setup_rec_from_negotiated_values(idbm_t *db, node_rec_t *rec,
+				struct session_info *info)
+{
+	struct iscsi_session_operational_config session_conf;
+	struct iscsi_conn_operational_config conn_conf;
+	struct iscsi_auth_config auth_conf;
+
+	idbm_node_setup_from_conf(db, rec);
+	strncpy(rec->name, info->targetname, TARGET_NAME_MAXLEN);
+	rec->conn[0].port = info->persistent_port;
+	strncpy(rec->conn[0].address, info->persistent_address, NI_MAXHOST);
+	memcpy(&rec->iface, &info->iface, sizeof(struct iface_rec));
+	rec->tpgt = info->tpgt;
+	iface_copy(&rec->iface, &info->iface);
+
+	get_negotiated_session_conf(info->sid, &session_conf);
+	get_negotiated_conn_conf(info->sid, &conn_conf);
+	get_auth_conf(info->sid, &auth_conf);
+
+	if (strlen(auth_conf.username))
+		strcpy(rec->session.auth.username, auth_conf.username);
+
+	if (strlen(auth_conf.username_in))
+		strcpy(rec->session.auth.username_in, auth_conf.username_in);
+
+	if (strlen((char *)auth_conf.password)) {
+		strcpy((char *)rec->session.auth.password,
+			(char *)auth_conf.password);
+		rec->session.auth.password_length = auth_conf.password_length;
+	}
+
+	if (strlen((char *)auth_conf.password_in)) {
+		strcpy((char *)rec->session.auth.password_in,
+			(char *)auth_conf.password_in);
+		rec->session.auth.password_in_length =
+						auth_conf.password_in_length;
+	}
+
+	if (is_valid_operational_value(conn_conf.HeaderDigest)) {
+		if (conn_conf.HeaderDigest)
+			rec->conn[0].iscsi.HeaderDigest =
+						CONFIG_DIGEST_PREFER_ON;
+		else
+			rec->conn[0].iscsi.HeaderDigest =
+						CONFIG_DIGEST_PREFER_OFF;
+	}
+
+	if (is_valid_operational_value(conn_conf.DataDigest)) {
+		if (conn_conf.DataDigest)
+			rec->conn[0].iscsi.DataDigest = CONFIG_DIGEST_PREFER_ON;
+		else
+			rec->conn[0].iscsi.DataDigest =
+						CONFIG_DIGEST_PREFER_OFF;
+	}
+
+	if (is_valid_operational_value(conn_conf.MaxRecvDataSegmentLength))
+		rec->conn[0].iscsi.MaxRecvDataSegmentLength =
+					conn_conf.MaxRecvDataSegmentLength;
+
+	if (is_valid_operational_value(conn_conf.MaxXmitDataSegmentLength))
+		 rec->conn[0].iscsi.MaxXmitDataSegmentLength =
+					conn_conf.MaxXmitDataSegmentLength;
+
+	if (is_valid_operational_value(session_conf.FirstBurstLength))
+		rec->session.iscsi.FirstBurstLength =
+					session_conf.FirstBurstLength;
+
+	if (is_valid_operational_value(session_conf.MaxBurstLength))
+		rec->session.iscsi.MaxBurstLength =
+					session_conf.MaxBurstLength;
+
+	if (is_valid_operational_value(session_conf.ImmediateData)) {
+		if (session_conf.ImmediateData)
+			rec->session.iscsi.ImmediateData = 1;
+		else
+			rec->session.iscsi.ImmediateData = 0;
+	}
+
+	if (is_valid_operational_value(session_conf.InitialR2T)) {
+		if (session_conf.InitialR2T)
+			rec->session.iscsi.InitialR2T = 0;
+		else
+			rec->session.iscsi.InitialR2T = 1;
+	}
+}
+
+static int sync_session(void *data, struct session_info *info)
 {
 	idbm_t *db = data;
 	node_rec_t rec;
-	int fd = -1;
 	iscsiadm_req_t req;
 	iscsiadm_rsp_t rsp;
-	iscsi_provider_t *p;
+	struct iscsi_transport *t;
 
-	log_debug(7, "sync session [%d][%s,%s.%d]\n", sid, targetname, address,
-		  port);
+	log_debug(7, "sync session [%d][%s,%s.%d][%s]\n", info->sid,
+		  info->targetname, info->persistent_address,
+		  info->port, info->iface.hwaddress);
 
-	p = get_transport_by_sid(sid);
-	if (!p)
+	t = get_transport_by_sid(info->sid);
+	if (!t)
 		return 0;
 
-	if (idbm_node_read(db, &rec, targetname, address, port)) {
-		log_warning("could not read data for [%s,%s.%d]\n",
-			    targetname, address, port);
+	/*
+	 * Just rescan the device in case this is the first startup.
+	 * (TODO: should do this async and check for state).
+	 */
+	if (t->caps & CAP_FW_DB) {
+		uint32_t host_no;
+		int err;
+
+		host_no = get_host_no_from_sid(info->sid, &err);
+		if (err) {
+			log_error("Could not get host no from sid %u. Can not "
+				  "sync session. Error %d", info->sid, err);
+			return 0;
+		}
+		scan_host(host_no, 0);
 		return 0;
 	}
 
-	/*
-	 * for now skip qlogic and other HW and offload driver that
-	 * that store their data somewhere else
-	 */
-	if (strcmp(p->name, rec.transport_name))
-		return 0;
+	memset(&rec, 0, sizeof(node_rec_t));
+	if (iface_get_by_bind_info(db, &info->iface, &rec.iface)) {
+		log_warning("Could not read data from db. Using default and "
+			    "currently negotiated values\n");
+		setup_rec_from_negotiated_values(db, &rec, info);
+	} else if (idbm_rec_read(db, &rec, info->targetname, info->tpgt,
+			  info->persistent_address, info->persistent_port,
+			  &rec.iface)) {
+		log_warning("Could not read data from db. Using default and "
+			    "currently negotiated values\n");
+		setup_rec_from_negotiated_values(db, &rec, info);
+	}
+
+	/* multiple drivers could be connected to the same portal */
+	if (!iscsi_match_session(&rec, info))
+		return -1;
 
 	memset(&req, 0, sizeof(req));
 	req.command = MGMT_IPC_SESSION_SYNC;
-	req.u.session.sid = sid;
+	req.u.session.sid = info->sid;
 	memcpy(&req.u.session.rec, &rec, sizeof(node_rec_t));
 
-	do_iscsid(&fd, &req, &rsp);
+	do_iscsid(&req, &rsp);
 	return 0;
 }
 
@@ -143,10 +253,23 @@ static void sync_sessions(void)
 	idbm_terminate(db);
 }
 
-static void catch_signal(int signo) {
+static void catch_signal(int signo)
+{
 	log_warning("caught signal -%d, ignoring...", signo);
 }
 
+static void iscsid_exit(void)
+{
+	log_debug(1, "iscsid_exit");
+	if (daemon_config.initiator_name)
+		free(daemon_config.initiator_name);
+	if (daemon_config.initiator_alias)
+		free(daemon_config.initiator_alias);
+	free_initiator();
+	mgmt_ipc_close(mgmt_ipc_fd);
+	ipc->ctldev_close();
+}
+
 int main(int argc, char *argv[])
 {
 	struct utsname host_info; /* will use to compound initiator alias */
@@ -154,6 +277,7 @@ int main(int argc, char *argv[])
 	char *initiatorname_file = INITIATOR_NAME_FILE;
 	char *pid_file = PID_FILE;
 	int ch, longindex;
+	int isns_fd;
 	uid_t uid = 0;
 	gid_t gid = 0;
 	struct sigaction sa_old;
@@ -207,46 +331,61 @@ int main(int argc, char *argv[])
 
 	/* make sure we have initiatorname config file first */
 	if (access(initiatorname_file, R_OK)) {
-		fprintf(stderr, "Error: initiatorname file doesn't exist!");
-		fprintf(stderr, " default [%s]\n", INITIATOR_NAME_FILE);
+		fprintf(stderr, "Error: initiatorname file %s doesn't exist.\n",
+			initiatorname_file);
+		fprintf(stderr, "Please create a file %s that contains\n",
+			initiatorname_file);
+		fprintf(stderr, "a sting with the format: InitiatorName="
+			"iqn.yyyy-mm.<reversed domain name>[:identifier].\n");
+		fprintf(stderr, "Example: InitiatorName=iqn.2001-04.com.redhat:"
+			"fc6\n");
 		usage(0);
 	}
 
 	/* initialize logger */
-	log_init(program_name, DEFAULT_AREA_SIZE);
+	log_pid = log_init(program_name, DEFAULT_AREA_SIZE);
+	if (log_pid < 0)
+		exit(1);
 	check_class_version();
 
 	umask(0177);
 
-	if ((mgmt_ipc_fd = mgmt_ipc_listen()) < 0) {
-		exit(-1);
+	mgmt_ipc_fd = -1;
+	control_fd = -1;
+	daemon_config.initiator_name = NULL;
+	daemon_config.initiator_alias = NULL;
+	if (atexit(iscsid_exit)) {
+		log_error("failed to set exit function\n");
+		exit(1);
 	}
 
+	if ((mgmt_ipc_fd = mgmt_ipc_listen()) < 0)
+		exit(-1);
+
 	if (log_daemon) {
 		char buf[64];
 		int fd;
 
 		fd = open(pid_file, O_WRONLY|O_CREAT, 0644);
 		if (fd < 0) {
-			log_error("unable to create pid file");
+			log_error("Unable to create pid file");
 			exit(1);
 		}
 		pid = fork();
 		if (pid < 0) {
-			log_error("starting daemon failed");
+			log_error("Starting daemon failed");
 			exit(1);
 		} else if (pid) {
-			log_warning("iSCSI daemon with pid=%d started!", pid);
+			log_error("iSCSI daemon with pid=%d started!", pid);
 			exit(0);
 		}
 
-		if ((control_fd = ipc->ctldev_open()) < 0) {
+		if ((control_fd = ipc->ctldev_open()) < 0)
 			exit(-1);
-		}
 
 		chdir("/");
 		if (lockf(fd, F_TLOCK, 0) < 0) {
-			log_error("unable to lock pid file");
+			log_error("Unable to lock pid file");
 			exit(1);
 		}
 		ftruncate(fd, 0);
@@ -255,9 +394,8 @@ int main(int argc, char *argv[])
 
 		daemon_init();
 	} else {
-		if ((control_fd = ipc->ctldev_open()) < 0) {
+		if ((control_fd = ipc->ctldev_open()) < 0)
 			exit(-1);
-		}
 	}
 
 	if (uid && setuid(uid) < 0)
@@ -291,7 +429,16 @@ int main(int argc, char *argv[])
 	log_debug(1, "InitiatorName=%s", daemon_config.initiator_name);
 	log_debug(1, "InitiatorAlias=%s", daemon_config.initiator_alias);
 
-	init_providers();
+	pid = fork();
+	if (pid == 0) {
+		/* child */
+		sync_sessions();
+		exit(0);
+	} else if (pid < 0) {
+		log_error("Fork failed error %d: existing sessions"
+			  " will not be synced", errno);
+	} else
+		need_reap();
 
 	/* oom-killer will not kill us at the night... */
 	if (oom_adjust())
@@ -303,22 +450,11 @@ int main(int argc, char *argv[])
 		exit(1);
 	}
 
-	pid = fork();
-	if (pid == 0) {
-		/* child */
-		sync_sessions();
-		exit(0);
-	} else if (pid < 0) {
-		log_error("fork failed error %d: existing sessions"
-			  " will not be synced", errno);
-	} else {
-		/* parent continues */
-		log_warning("iSCSI sync pid=%d started", pid);
-		need_reap();
-	}
-
 	actor_init();
-	event_loop(ipc, control_fd, mgmt_ipc_fd);
+	isns_fd = isns_init();
+	event_loop(ipc, control_fd, mgmt_ipc_fd, isns_fd);
+	isns_exit();
+
 	log_debug(1, "daemon stopping");
 	return 0;
 }
