@@ -41,6 +41,7 @@
 #include "idbm.h"
 #include "version.h"
 #include "iscsi_sysfs.h"
+#include "iscsi_settings.h"
 
 /* global config info */
 /* initiator needs initiator name/alias */
@@ -50,7 +51,6 @@ struct iscsi_daemon_config *dconfig = &daemon_config;
 static node_rec_t config_rec;
 
 static char program_name[] = "iscsistart";
-static int ipc_fd;
 static int mgmt_ipc_fd;
 
 /* used by initiator */
@@ -107,8 +107,8 @@ static int stop_event_loop(void)
 
 	memset(&req, 0, sizeof(req));
 	req.command = MGMT_IPC_IMMEDIATE_STOP;
-	rc = do_iscsid(&ipc_fd, &req, &rsp);
-	if (rc > 0)
+	rc = do_iscsid(&req, &rsp);
+	if (rc)
 		iscsid_handle_error(rc);
 	if (rc)
 		log_error("Could not stop event_loop\n");
@@ -122,11 +122,15 @@ static int setup_session(void)
 	iscsiadm_rsp_t rsp;
 	int rc;
 
+	/* we cannot answer so turn off */
+	config_rec.conn[0].timeo.noop_out_interval = 0;
+	config_rec.conn[0].timeo.noop_out_timeout = 0;
+
 	memset(&req, 0, sizeof(req));
 	req.command = MGMT_IPC_SESSION_LOGIN;
 	memcpy(&req.u.session.rec, &config_rec, sizeof(node_rec_t));
-	rc = do_iscsid(&ipc_fd, &req, &rsp);
-	if (rc > 0)
+	rc = do_iscsid(&req, &rsp);
+	if (rc)
 		iscsid_handle_error(rc);
 
 	return rc;
@@ -263,8 +267,6 @@ int main(int argc, char *argv[])
 	if (check_params(initiatorname))
 		exit(1);
 
-	init_providers();
-
 	pid = fork();
 	if (pid < 0) {
 		log_error("iscsiboot fork failed");
@@ -328,9 +330,10 @@ int main(int argc, char *argv[])
 	 * Start Main Event Loop
 	 */
 	actor_init();
-	event_loop(ipc, control_fd, mgmt_ipc_fd);
+	event_loop(ipc, control_fd, mgmt_ipc_fd, -1);
 	ipc->ctldev_close();
 	mgmt_ipc_close(mgmt_ipc_fd);
+	free_initiator();
 
 	log_debug(1, "iscsi child done");
 	return 0;
