@@ -54,7 +54,7 @@ mgmt_ipc_listen(void)
 
 	fd = socket(AF_LOCAL, SOCK_STREAM, 0);
 	if (fd < 0) {
-		log_error("can not create IPC socket");
+		log_error("Can not create IPC socket");
 		return fd;
 	}
 
@@ -64,19 +64,17 @@ mgmt_ipc_listen(void)
 		strlen(ISCSIADM_NAMESPACE));
 
 	if ((err = bind(fd, (struct sockaddr *) &addr, sizeof(addr))) < 0) {
-		log_error("can not bind IPC socket");
+		log_error("Can not bind IPC socket");
 		close(fd);
 		return err;
 	}
 
 	if ((err = listen(fd, 32)) < 0) {
-		log_error("can not listen IPC socket");
+		log_error("Can not listen IPC socket");
 		close(fd);
 		return err;
 	}
 
-	log_debug(1, "IPC socket is listening...");
-
 	return fd;
 }
 
@@ -101,21 +99,15 @@ static mgmt_ipc_err_e
 mgmt_ipc_session_getstats(queue_task_t *qtask, int sid,
 			  iscsiadm_rsp_t *rsp)
 {
-	iscsi_provider_t *p;
+	struct iscsi_transport *t;
 	iscsi_session_t *session;
-	struct qelem *sitem, *pitem;
 
-	pitem = providers.q_forw;
-	while (pitem != &providers) {
-		p = (iscsi_provider_t *)pitem;
-
-		sitem = p->sessions.q_forw;
-		while (sitem != &p->sessions) {
-			session = (iscsi_session_t *)sitem;
+	list_for_each_entry(t, &transports, list) {
+		list_for_each_entry(session, &t->sessions, list) {
 			if (session->id == sid) {
 				int rc;
 
-				rc = ipc->get_stats(session->transport_handle,
+				rc = ipc->get_stats(session->t->handle,
 					session->id, session->conn[0].id,
 					(void*)&rsp->u.getstats,
 					MGMT_IPC_GETSTATS_BUF_MAX);
@@ -126,9 +118,7 @@ mgmt_ipc_session_getstats(queue_task_t *qtask, int sid,
 				}
 				return MGMT_IPC_OK;
 			}
-			sitem = sitem->q_forw;
 		}
-		pitem = pitem->q_forw;
 	}
 
 	return MGMT_IPC_ERR_NOT_FOUND;
@@ -166,7 +156,7 @@ static mgmt_ipc_err_e
 mgmt_ipc_session_info(queue_task_t *qtask, int sid, iscsiadm_rsp_t *rsp)
 {
 	iscsi_session_t *session;
-	struct msg_session_state *info;
+	struct ipc_msg_session_state *info;
 
 	if (!(session = session_find_by_sid(sid))) {
 		log_error("session with sid %d not found!", sid);
@@ -207,6 +197,12 @@ mgmt_ipc_conn_remove(queue_task_t *qtask, int cid)
 	return MGMT_IPC_ERR;
 }
 
+static mgmt_ipc_err_e
+mgmt_ipc_isns_dev_attr_query(queue_task_t *qtask)
+{
+	return isns_dev_attr_query_task(qtask);
+}
+
 static int
 mgmt_peeruser(int sock, char *user)
 {
@@ -291,6 +287,23 @@ mgmt_peeruser(int sock, char *user)
 #endif
 }
 
+void
+mgmt_ipc_write_rsp(queue_task_t *qtask, mgmt_ipc_err_e err)
+{
+	if (!qtask)
+		return;
+	log_debug(4, "%s: rsp to fd %d", __FUNCTION__,
+		 qtask->mgmt_ipc_fd);
+
+	if (qtask->mgmt_ipc_fd < 0)
+		return;
+
+	qtask->rsp.err = err;
+	write(qtask->mgmt_ipc_fd, &qtask->rsp, sizeof(qtask->rsp));
+	close(qtask->mgmt_ipc_fd);
+	free(qtask);
+}
+
 static int
 mgmt_ipc_handle(int accept_fd)
 {
@@ -312,8 +325,7 @@ mgmt_ipc_handle(int accept_fd)
 		return rc;
 	}
 
-	if (!mgmt_peeruser(accept_fd, user) ||
-	    strncmp(user, "root", PEERUSER_MAX)) {
+	if (!mgmt_peeruser(fd, user) || strncmp(user, "root", PEERUSER_MAX)) {
 		rsp.err = MGMT_IPC_ERR_ACCESS;
 		rc = EINVAL;
 		goto err;
@@ -332,8 +344,8 @@ mgmt_ipc_handle(int accept_fd)
 		rc = -ENOMEM;
 		goto err;
 	}
-	memcpy(&qtask->u.login.req, &req, sizeof(iscsiadm_req_t));
-	qtask->u.login.mgmt_ipc_fd = fd;
+	memcpy(&qtask->req, &req, sizeof(iscsiadm_req_t));
+	qtask->mgmt_ipc_fd = fd;
 
 	switch(req.command) {
 	case MGMT_IPC_SESSION_LOGIN:
@@ -351,6 +363,12 @@ mgmt_ipc_handle(int accept_fd)
 						    &rsp);
 		immrsp = 1;
 		break;
+	case MGMT_IPC_SEND_TARGETS:
+		rsp.err = iscsi_host_send_targets(qtask, req.u.st.host_no,
+						  req.u.st.do_login,
+						  &req.u.st.ss);
+		immrsp = 1;
+		break;
 	case MGMT_IPC_SESSION_INFO:
 		rsp.err = mgmt_ipc_session_info(qtask, req.u.session.sid,
 						&rsp);
@@ -379,6 +397,15 @@ mgmt_ipc_handle(int accept_fd)
 		immrsp = 1;
 		rc = 1;
 		break;
+	case MGMT_IPC_ISNS_DEV_ATTR_QUERY:
+		rsp.err = mgmt_ipc_isns_dev_attr_query(qtask);
+		break;
+	case MGMT_IPC_SET_HOST_PARAM:
+		rsp.err = iscsi_host_set_param(req.u.set_host_param.host_no,
+						req.u.set_host_param.param,
+						req.u.set_host_param.value);
+		immrsp = 1;
+		break;
 	default:
 		log_error("unknown request: %s(%d) %u",
 			  __FUNCTION__, __LINE__, req.command);
@@ -428,10 +455,12 @@ reaper(void)
 
 #define POLL_CTRL	0
 #define POLL_IPC	1
-#define POLL_MAX	2
+#define POLL_ISNS	2
+#define POLL_MAX	3
 
 /* TODO: this should go somewhere else */
-void event_loop(struct iscsi_ipc *ipc, int control_fd, int mgmt_ipc_fd)
+void event_loop(struct iscsi_ipc *ipc, int control_fd, int mgmt_ipc_fd,
+		int isns_fd)
 {
 	struct pollfd poll_array[POLL_MAX];
 	int res;
@@ -441,6 +470,13 @@ void event_loop(struct iscsi_ipc *ipc, int control_fd, int mgmt_ipc_fd)
 	poll_array[POLL_IPC].fd = mgmt_ipc_fd;
 	poll_array[POLL_IPC].events = POLLIN;
 
+	if (isns_fd < 0)
+		poll_array[POLL_ISNS].fd = poll_array[POLL_ISNS].events = 0;
+	else {
+		poll_array[POLL_ISNS].fd = isns_fd;
+		poll_array[POLL_ISNS].events = POLLIN;
+	}
+
 	while (1) {
 		res = poll(poll_array, POLL_MAX, ACTOR_RESOLUTION);
 		if (res > 0) {
@@ -451,6 +487,9 @@ void event_loop(struct iscsi_ipc *ipc, int control_fd, int mgmt_ipc_fd)
 			if (poll_array[POLL_IPC].revents)
 				if (mgmt_ipc_handle(mgmt_ipc_fd) == 1)
 					break;
+			if (poll_array[POLL_ISNS].revents)
+				isns_handle(isns_fd);
+
 		} else if (res < 0) {
 			if (errno == EINTR) {
 				log_debug(1, "event_loop interrupted");
