@@ -44,6 +44,8 @@ typedef enum mgmt_ipc_err {
 	MGMT_IPC_ERR_TRANS_CAPS		= 14,
 	MGMT_IPC_ERR_EXISTS		= 15,
 	MGMT_IPC_ERR_INVALID_REQ	= 16,
+	MGMT_IPC_ERR_ISNS_UNAVAILABLE	= 17,
+	MGMT_IPC_ERR_ISCSID_COMM_ERR	= 18,
 } mgmt_ipc_err_e;
 
 typedef enum iscsiadm_cmd {
@@ -60,6 +62,9 @@ typedef enum iscsiadm_cmd {
 	MGMT_IPC_IMMEDIATE_STOP		= 11,
 	MGMT_IPC_SESSION_SYNC		= 12,
 	MGMT_IPC_SESSION_INFO		= 13,
+	MGMT_IPC_ISNS_DEV_ATTR_QUERY	= 14,
+	MGMT_IPC_SEND_TARGETS		= 15,
+	MGMT_IPC_SET_HOST_PARAM		= 16,
 } iscsiadm_cmd_e;
 
 typedef enum iscsi_conn_state_e {
@@ -85,14 +90,26 @@ typedef struct iscsiadm_req {
 
 	union {
 		/* messages */
-		struct msg_session {
+		struct ipc_msg_session {
 			int sid;
 			node_rec_t rec;
 		} session;
-		struct msg_conn {
+		struct ipc_msg_conn {
 			int sid;
 			int cid;
 		} conn;
+		struct ipc_msg_send_targets {
+			int host_no;
+			int do_login;
+			struct sockaddr_storage ss;
+		} st;
+		struct ipc_msg_set_host_param {
+			int host_no;
+			int param;
+			/* TODO: make this variable len to support */
+			char value[IFNAMSIZ + 1];
+
+		} set_host_param;
 	} u;
 } iscsiadm_req_t;
 
@@ -106,16 +123,16 @@ typedef struct iscsiadm_rsp {
 					sizeof(struct iscsi_stats) + \
 					sizeof(struct iscsi_stats_custom) * \
 						ISCSI_STATS_CUSTOM_MAX)
-		struct msg_getstats {
+		struct ipc_msg_getstats {
 			struct iscsi_uevent ev;
 			struct iscsi_stats stats;
 			char custom[sizeof(struct iscsi_stats_custom) *
 					ISCSI_STATS_CUSTOM_MAX];
 		} getstats;
-		struct msg_config {
+		struct ipc_msg_config {
 			char var[VALUE_MAXLEN];
 		} config;
-		struct msg_session_state {
+		struct ipc_msg_session_state {
 			iscsi_session_r_stage_e session_state;
 			iscsi_conn_state_e conn_state;
 		} session_state;
@@ -125,7 +142,10 @@ typedef struct iscsiadm_rsp {
 struct iscsi_ipc *ipc;
 
 void need_reap(void);
-void event_loop(struct iscsi_ipc *ipc, int control_fd, int mgmt_ipc_fd);
+void event_loop(struct iscsi_ipc *ipc, int control_fd, int mgmt_ipc_fd, int isns_fd);
+
+struct queue_task;
+void mgmt_ipc_write_rsp(struct queue_task *qtask, mgmt_ipc_err_e err);
 int mgmt_ipc_listen(void);
 void mgmt_ipc_close(int fd);
 
