@@ -37,6 +37,7 @@
 #include "iscsi_ipc.h"
 #include "initiator.h"
 #include "iscsi_sysfs.h"
+#include "transport.h"
 
 static int ctrl_fd;
 static struct sockaddr_nl src_addr, dest_addr;
@@ -52,11 +53,11 @@ static void *setparam_buf;
 static int ctldev_handle(void);
 
 #define NLM_BUF_DEFAULT_MAX \
-	(NLMSG_SPACE(DEFAULT_MAX_RECV_DATA_SEGMENT_LENGTH + \
+	(NLMSG_SPACE(ISCSI_DEF_MAX_RECV_SEG_LEN + \
 			 sizeof(struct iscsi_hdr)))
 
 #define PDU_SENDBUF_DEFAULT_MAX \
-	(DEFAULT_MAX_RECV_DATA_SEGMENT_LENGTH + sizeof(struct iscsi_hdr))
+	(ISCSI_DEF_MAX_RECV_SEG_LEN + sizeof(struct iscsi_hdr))
 
 #define NLM_SETPARAM_DEFAULT_MAX \
 	(NI_MAXHOST + 1 + sizeof(struct iscsi_uevent))
@@ -303,8 +304,43 @@ __kipc_call(void *iov_base, int iov_len)
 	return rc;
 }
 
+static int
+ksendtargets(uint64_t transport_handle, uint32_t host_no, struct sockaddr *addr)
+{
+	int rc, addrlen;
+	struct iscsi_uevent *ev;
+
+	log_debug(7, "in %s", __FUNCTION__);
+
+	memset(setparam_buf, 0, NLM_SETPARAM_DEFAULT_MAX);
+	ev = (struct iscsi_uevent *)setparam_buf;
+	ev->type = ISCSI_UEVENT_TGT_DSCVR;
+	ev->transport_handle = transport_handle;
+	ev->u.tgt_dscvr.type = ISCSI_TGT_DSCVR_SEND_TARGETS;
+	ev->u.tgt_dscvr.host_no = host_no;
+
+	if (addr->sa_family == PF_INET)
+		addrlen = sizeof(struct sockaddr_in);
+	else if (addr->sa_family == PF_INET6)
+		addrlen = sizeof(struct sockaddr_in6);
+	else {
+		log_error("%s unknown addr family %d\n",
+			  __FUNCTION__, addr->sa_family);
+		return -EINVAL;
+	}
+	memcpy(setparam_buf + sizeof(*ev), addr, addrlen);
+
+	rc = __kipc_call(ev, sizeof(*ev) + addrlen);
+	if (rc < 0) {
+		log_error("sendtargets failed rc%d\n", rc);
+		return rc;
+	}
+	return 0;
+}
+
 static int
 kcreate_session(uint64_t transport_handle, uint32_t initial_cmdsn,
+		uint16_t cmds_max, uint16_t qdepth,
 		uint32_t *out_sid, uint32_t *out_hostno)
 {
 	int rc;
@@ -317,6 +353,8 @@ kcreate_session(uint64_t transport_handle, uint32_t initial_cmdsn,
 	ev.type = ISCSI_UEVENT_CREATE_SESSION;
 	ev.transport_handle = transport_handle;
 	ev.u.c_session.initial_cmdsn = initial_cmdsn;
+	ev.u.c_session.cmds_max = cmds_max;
+	ev.u.c_session.queue_depth = qdepth;
 
 	if ((rc = __kipc_call(&ev, sizeof(ev))) < 0) {
 		return rc;
@@ -501,6 +539,46 @@ err:
 	return rc;
 }
 
+static int
+kset_host_param(uint64_t transport_handle, uint32_t host_no,
+		enum iscsi_host_param param, void *value, int type)
+{
+	struct iscsi_uevent *ev;
+	char *param_str;
+	int rc, len;
+
+	log_debug(7, "in %s", __FUNCTION__);
+
+	memset(setparam_buf, 0, NLM_SETPARAM_DEFAULT_MAX);
+	ev = (struct iscsi_uevent *)setparam_buf;
+	ev->type = ISCSI_UEVENT_SET_HOST_PARAM;
+	ev->transport_handle = transport_handle;
+	ev->u.set_host_param.host_no = host_no;
+	ev->u.set_host_param.param = param;
+
+	param_str = setparam_buf + sizeof(*ev);
+	switch (type) {
+	case ISCSI_INT:
+		sprintf(param_str, "%d", *((int *)value));
+		break;
+	case ISCSI_STRING:
+		if (!strlen(value))
+			return 0;
+		sprintf(param_str, "%s", (char *)value);
+		break;
+	default:
+		log_error("invalid type %d\n", type);
+		return -EINVAL;
+	}
+	ev->u.set_host_param.len = len = strlen(param_str) + 1;
+
+	if ((rc = __kipc_call(ev, sizeof(*ev) + len)) < 0) {
+		return rc;
+	}
+
+	return 0;
+}
+
 static int
 kset_param(uint64_t transport_handle, uint32_t sid, uint32_t cid,
 	   enum iscsi_param param, void *value, int type)
@@ -525,6 +603,8 @@ kset_param(uint64_t transport_handle, uint32_t sid, uint32_t cid,
 		sprintf(param_str, "%d", *((int *)value));
 		break;
 	case ISCSI_STRING:
+		if (!strlen(value))
+			return 0;
 		sprintf(param_str, "%s", (char *)value);
 		break;
 	default:
@@ -638,7 +718,7 @@ ktransport_ep_connect(iscsi_conn_t *conn, int non_blocking)
 	memset(setparam_buf, 0, NLM_SETPARAM_DEFAULT_MAX);
 	ev = (struct iscsi_uevent *)setparam_buf;
 	ev->type = ISCSI_UEVENT_TRANSPORT_EP_CONNECT;
-	ev->transport_handle = conn->session->transport_handle;
+	ev->transport_handle = conn->session->t->handle;
 
 	if (dst_addr->sa_family == PF_INET)
 		addrlen = sizeof(struct sockaddr_in);
@@ -676,7 +756,7 @@ ktransport_ep_poll(iscsi_conn_t *conn, int timeout_ms)
 	memset(&ev, 0, sizeof(struct iscsi_uevent));
 
 	ev.type = ISCSI_UEVENT_TRANSPORT_EP_POLL;
-	ev.transport_handle = conn->session->transport_handle;
+	ev.transport_handle = conn->session->t->handle;
 	ev.u.ep_poll.ep_handle  = conn->transport_ep_handle;
 	ev.u.ep_poll.timeout_ms = timeout_ms;
 
@@ -700,7 +780,7 @@ ktransport_ep_disconnect(iscsi_conn_t *conn)
 	memset(&ev, 0, sizeof(struct iscsi_uevent));
 
 	ev.type = ISCSI_UEVENT_TRANSPORT_EP_DISCONNECT;
-	ev.transport_handle = conn->session->transport_handle;
+	ev.transport_handle = conn->session->t->handle;
 	ev.u.ep_disconnect.ep_handle = conn->transport_ep_handle;
 
 	if ((rc = __kipc_call(&ev, sizeof(ev))) < 0) {
@@ -757,13 +837,19 @@ kget_stats(uint64_t transport_handle, uint32_t sid, uint32_t cid,
 	return 0;
 }
 
-static int
-ctldev_handle(void)
+static void drop_data(struct nlmsghdr *nlh)
+{
+	int ev_size;
+
+	ev_size = nlh->nlmsg_len - NLMSG_ALIGN(sizeof(struct nlmsghdr));
+	nlpayload_read(ctrl_fd, setparam_buf, ev_size, 0);
+}
+
+static int ctldev_handle(void)
 {
 	int rc;
 	struct iscsi_uevent *ev;
-	struct qelem *item, *pitem;
-	iscsi_provider_t *p;
+	struct iscsi_transport *t;
 	iscsi_session_t *session = NULL;
 	iscsi_conn_t *conn = NULL;
 	uintptr_t recv_handle;
@@ -781,15 +867,28 @@ ctldev_handle(void)
 	nlh = (struct nlmsghdr *)nlm_ev;
 	ev = (struct iscsi_uevent *)NLMSG_DATA(nlm_ev);
 
-	/* verify connection */
-	pitem = providers.q_forw;
-	while (pitem != &providers) {
-		p = (iscsi_provider_t *)pitem;
+	log_debug(7, "%s got event type %u\n", __FUNCTION__, ev->type);
+	/* drivers like qla4xxx can be inserted after iscsid is started */
+	switch (ev->type) {
+	case ISCSI_UEVENT_CREATE_SESSION:
+		drop_data(nlh);
+		iscsi_async_session_creation(ev->r.c_session_ret.host_no,
+					     ev->r.c_session_ret.sid);
+		return 0;
+	case ISCSI_KEVENT_DESTROY_SESSION:
+		drop_data(nlh);
+		iscsi_async_session_destruction(ev->r.d_session.host_no,
+						ev->r.d_session.sid);
+		return 0;
+	default:
+		; /* fall through */
+	}
 
-		item = p->sessions.q_forw;
-		while (item != &p->sessions) {
+	/* verify connection */
+	list_for_each_entry(t, &transports, list) {
+		list_for_each_entry(session, &t->sessions, list) {
 			int i;
-			session = (iscsi_session_t *)item;
+
 			for (i=0; i<ISCSI_CONN_MAX; i++) {
 				if (ev->type == ISCSI_KEVENT_RECV_PDU &&
 				    session->id == ev->r.recv_req.sid &&
@@ -804,22 +903,22 @@ ctldev_handle(void)
 					goto verify_conn;
 				}
 			}
-			item = item->q_forw;
 		}
-		pitem = pitem->q_forw;
 	}
 
 verify_conn:
 	if (conn == NULL) {
-		log_error("could not verify connection %d:%d",
-			  ev->r.recv_req.sid, ev->r.recv_req.cid);
+		log_error("Could not verify connection %d:%d. Dropping "
+			   "event.\n", ev->r.recv_req.sid, ev->r.recv_req.cid);
+		drop_data(nlh);
 		return -ENXIO;
 	}
 
 	ev_size = nlh->nlmsg_len - NLMSG_ALIGN(sizeof(struct nlmsghdr));
 	recv_handle = (uintptr_t)recvpool_get(conn, ev_size);
 	if (!recv_handle) {
-		log_error("can not allocate memory for receive handle");
+		/* retry later */
+		log_error("Can not allocate memory for receive handle.");
 		return -ENOMEM;
 	}
 
@@ -830,21 +929,26 @@ verify_conn:
 				ev_size, 0)) < 0) {
 		recvpool_put(conn, (void*)recv_handle);
 		log_error("can not read from NL socket, error %d", rc);
+		/* retry later */
 		return rc;
 	}
 
-	if (ev->type == ISCSI_KEVENT_RECV_PDU) {
+	switch (ev->type) {
+	case ISCSI_KEVENT_RECV_PDU:
 		/* produce an event, so session manager will handle */
 		queue_produce(session->queue, EV_CONN_RECV_PDU, conn,
 			sizeof(uintptr_t), &recv_handle);
 		actor_schedule(&session->mainloop);
-	} else if (ev->type == ISCSI_KEVENT_CONN_ERROR) {
+		break;
+	case ISCSI_KEVENT_CONN_ERROR:
 		/* produce an event, so session manager will handle */
+		memcpy((void *)recv_handle, &ev->r.connerror.error,
+			sizeof(ev->r.connerror.error));
 		queue_produce(session->queue, EV_CONN_ERROR, conn,
-			sizeof(uintptr_t), (void*)&ev->r.connerror.error);
+			sizeof(uintptr_t), &recv_handle);
 		actor_schedule(&session->mainloop);
-		recvpool_put(conn, (void*)recv_handle);
-	} else {
+		break;
+	default:
 		recvpool_put(conn, (void*)recv_handle);
 		log_error("unknown kernel event %d", ev->type);
 		return -EEXIST;
@@ -883,7 +987,7 @@ ctldev_open(void)
 	}
 
 	ctrl_fd = socket(PF_NETLINK, SOCK_RAW, NETLINK_ISCSI);
-	if (!ctrl_fd) {
+	if (ctrl_fd < 0) {
 		log_error("can not create NETLINK_ISCSI socket");
 		goto free_setparam_buf;
 	}
@@ -928,7 +1032,8 @@ ctldev_close(void)
 	free(pdu_sendbuf);
 	free(nlm_recvbuf);
 	free(nlm_sendbuf);
-	close(ctrl_fd);
+	if (ctrl_fd >= 0)
+		close(ctrl_fd);
 }
 
 struct iscsi_ipc nl_ipc = {
@@ -937,12 +1042,14 @@ struct iscsi_ipc nl_ipc = {
 	.ctldev_open		= ctldev_open,
 	.ctldev_close		= ctldev_close,
 	.ctldev_handle		= ctldev_handle,
+	.sendtargets		= ksendtargets,
 	.create_session         = kcreate_session,
 	.destroy_session        = kdestroy_session,
 	.create_conn            = kcreate_conn,
 	.destroy_conn           = kdestroy_conn,
 	.bind_conn              = kbind_conn,
 	.set_param              = kset_param,
+	.set_host_param		= kset_host_param,
 	.get_param              = NULL,
 	.start_conn             = kstart_conn,
 	.stop_conn              = kstop_conn,
