@@ -18,11 +18,12 @@
 #define ISCSI_TRANSPORT_H
 
 #include "types.h"
+#include "config.h"
 
-struct iscsi_provider_t;
+struct iscsi_transport;
 struct iscsi_conn;
 
-struct iscsi_uspace_transport {
+struct iscsi_transport_template {
 	const char *name;
 	uint8_t rdma;
 	int (*ep_connect) (iscsi_conn_t *conn, int non_blocking);
@@ -30,6 +31,16 @@ struct iscsi_uspace_transport {
 	void (*ep_disconnect) (iscsi_conn_t *conn);
 };
 
-extern int set_uspace_transport(struct iscsi_provider_t *p);
+/* represents data path provider */
+struct iscsi_transport {
+	struct list_head list;
+	uint64_t handle;
+	uint32_t caps;
+	char name[ISCSI_TRANSPORT_NAME_MAXLEN];
+	struct list_head sessions;
+	struct iscsi_transport_template *template;
+};
+
+extern int set_transport_template(struct iscsi_transport *t);
 
 #endif
